#!/usr/bin/perl -w

=head1 NAME

Eidetic - Templatized web-based database viewer, editor, indexer, etc

=head1 SYNOPSIS

Eidetic uses a database schema and a collection of Template Toolkit
templates to present webpages.  These pages can be used to examine or
update data, and can have subroutines hooked to them for security,
specialized updates, etc.

=head1 DESCRIPTION

Eidetic can be grouped into four aspects:

 - Database
 - Queries
 - Templates
 - Code

The Eidetic Database has several rules about how queries are built, and
how fields are used.  The field "rsf" (resource flag) must exist.  All
rows with rsf!=1 are ignored during queries.  All tables must have an
indexed column, whose name is shared by all tables.  For example, the
"user" table is indexed by "uid", the "host" table is indexed by "uid".
All junction tables are named after the two tables that are joined.  For
example, the "user_to_host" table is used to find which "host" entries
are tied to which "user" entries.  Field names within a junction table
must have the originating table name and the indexed column name.  For
example, the above junction table would have "uid" to identify the
junction row, along with "user_uid" and "host_uid" for the junction
itself.

Eidetic Queries are used to populate Templates.  They are defined either
globally (available to all modules), or defined per-module.  Queries
contain the table, filter, and sorting order.  EXAMPLE NEEDED

Eidetic Templates are Template Toolkit templates, and are processed with
all the database fields of the Query populated.  Several Eidetic
reserved-words are also populated.  These are:

  TO DOCUMENT

Eidetic Code are extra pieces of code outside the core Eidetic functions,
used to process Eidetic commands.  For example, if not every user is allowed
to update a certain template, there must be code associated with the
module that will check their permissions.  EXAMPLE NEEDED


=head1 FUNCTIONS

=cut

package Eidetic;

use strict;
use vars qw($VERSION @ISA); # Standard object stuff

#                    M mmmrrr M=Major m=minor r=revision (even minor=stable)
our $VERSION       = 2.003_003;

use fields qw(
         _GLOBALS
         _CONFIG
         _ARGS

         _user
         _cgi
         _subobj

         _tables
         _DB
         _AUTH_DB
         _db_handles

         _modules
         _queries

         _idField

         _storageError
         _eideticDefaults

         _template

         Error

         _TRACE
         _DEBUG
);
use vars qw( %FIELDS );

# Standard modules from CPAN
use CGI;
use IO::File;
use POSIX                qw(tmpnam strftime);
use Digest::MD5          qw(md5_hex);
use Text::Wrap           qw(40 &wrap);
use File::Copy;
use File::Basename;     # fileparse()
use File::Spec::Functions qw(:ALL);
use Template;
use Sys::Hostname;
use Carp                 qw(carp confess);
use Config::Simple;
use Sort::Tree;
use Crypt::SmbHash qw(nthash);

# Modules for this application
use Eidetic::Utility     qw(parseOn loadFile trace_tree generateRandomPassword);
use Eidetic::Database;

# Modules for the project application
use Mail::Sender;

=head2 new

Creates a new Eidetic Object.

 new(
      program_name => 'name of the application (defaults to $0)',
      idField => 'name of the joining id field in the db (defaults to "uid")'
      DEBUG => debug level.  0 = off, 3 = tons
      TRACE => function call tracing on or off.
     )

=cut

sub new
{
    my ($this,%args) = @_;
    my $class = ref($this) || $this;
    my $self = bless [\%FIELDS], $class;

    $self->{_cgi}           = new CGI;

    %{ $self->{_GLOBALS} } = (
	       operating_system     => $^O                    || '',
	       process_id           => $$                     || '',
	       process_user_id      => scalar getpwuid($>)    || '',
	       debugging            => $^D                    || '',
	       program_name         => $args{'program_name'}  || $0,
	       perl_version         => $]                     || '',
	       executable_name      => $^X                    || '',

	       eidetic_version      => $VERSION               || '',

           # These should go away?
           #
           # The idea is that we pass this GLOBALS structure in to
           # the TT2 processor, and these seemed like variables that
           # could be of usefulness to the template author - bwh
           #
           # But maybe they can be accessed from the $self->{_cgi} object -- bwh
           #
	       self_url             => $self->{_cgi}->self_url         || '',
	       query_string         => $self->{_cgi}->query_string     || '',
	       user_agent           => $self->{_cgi}->user_agent()     || '',
	       path_info            => $self->{_cgi}->path_info()      || '',
	       remote_host          => $self->{_cgi}->remote_host()    || '',
	       script_name          => $self->{_cgi}->script_name()    || '',
	       referer              => $self->{_cgi}->referer()        || '',
	       auth_type            => $self->{_cgi}->auth_type()      || '',
	       server_name          => $self->{_cgi}->server_name()    || '',
	       virtual_host         => $self->{_cgi}->virtual_host()   || '',
	       server_port          => $self->{_cgi}->server_port()    || '',
	       remote_user          => $self->{_cgi}->remote_user()    || '',
	       request_method       => $self->{_cgi}->request_method() || '',
	       content_type         => $self->{_cgi}->content_type()   || '',
	);
    $self->{_idField}     = $args{'idField'} || 'uid';
    $self->{_DEBUG}       = $args{'DEBUG'}   || 0;
    $self->{_TRACE}       = defined($args{'TRACE'}) ? 1 : 0;

    if ($self->{_DEBUG}>3) {
        warn "-" x 72, "\n";
        warn "EIDETIC GLOBALS\n";
        warn "-" x 72, "\n";
   
        while (my ($k,$v) = each %{ $self->{_GLOBALS} }) {
           	warn $k, ' 'x(25-length($k)), " = ", $v, "\n";
        }
        warn "-" x 72, "\n";
    }


    # For tracking errors encountered when validating input
    %{$self->{_storageError}}=();


    # Used when building CGI commands... 
    # TODO: needs to be externally configurable
    %{$self->{_eideticDefaults}} = (
                       modulename => '',
                       command    => '',
                       on         => '',
                       start      => 0,
                       query      => '',
                       filter     => '',
                       sort       => '',
                       maxitems   => 25,
                       hilite     => '',
                       page       => 0
                       );

    return $self;
}

# Private tracing function.  Call this for reach function we enter.
# Records a full trace for CGI-based stuff, or real-time for cmdline.
sub trace
{
    my($self,$call)=@_;

    if ($self->{_TRACE}) {
	    if ($self->{_cgi}->server_software ne 'cmdline') {
            $self->{_GLOBALS}->{trace_log}.=$call."<BR>\n";
        } else {
            warn $call."\n";
        }
    }
}

# Does a trace dump, if one has been recorded
sub trace_dump
{
    my ($self)=@_;

    if ($self->{_TRACE} &&
        defined($self->{_GLOBALS}->{trace_log})) {
        print $self->{_GLOBALS}->{trace_log};
    }
}

=head2 configure(HASH)

Configures this instance of Eidetic.  Attempts to load
ETCDIR/program_name.conf to read configuration options.  Attempts to
start a DB connection using the fields "db_brand", "db_name", "db_host",
"db_user", and "db_pass".

In the "fields" hash, if a name has a value, that is the default value
used if the field is not found in the config file.  If the default
value is "undef", then that argument MUST be found in the config file.

Returns "undef" on failure, "1" on success.

 fields => {
                optional_config_field_name => DEFAULT_VALUE,
                required_config_field_name => undef,
           },

Default fields are:

 applicationTemplates:   root directory of all module templates
 globalHtmlRoot:         same as DocumentRoot in apache, used to locate files
 globalTemplates:        root directory for all site-wide templates
 globalHeader:           filename of header template (default "header.tmpl")
 globalFooter:           filename of footer template (default "footer.tmpl")
 webmaster:              email address of the webmaster for errors

=cut

# Bug/Feature: does not lose previous settings when called again
sub configure
{
    my ($self,%args) = @_;
    $self->trace("configure") if ($self->{_TRACE});

    # == Configuration ======================================================

    # Application name
    my ($NAME) = fileparse($self->{_GLOBALS}->{program_name}, qr{\.[^\.]+});
    if (!$NAME) {
        $self->{Error} = "Could not determine app name from $self->{_GLOBALS}->{program_name}\n";
        return undef;
    }

    my $CONFDIR = '%ETCDIR%';  # Filled in by make (see Makefile.PL)
    $CONFDIR="./t" unless ($CONFDIR ne '%ETC'.'DIR%');
    if (!-d $CONFDIR) {
        $self->{Error} = "Improper installation - invalid config dir '$CONFDIR'";
        return undef;
    }

    my $confname=catfile($CONFDIR, "$NAME.conf");
    if (!Config::Simple->import_from($confname, \%{$self->{_CONFIG}})) {
        $self->{Error} = "Config file not found: $confname: ".Config::Simple->error()."\n";
        return undef;
    }

    # Create database 
    $self->{_DB} = new Eidetic::Database(
                                         brand  => $self->{_CONFIG}->{db_brand},
                                         name   => $self->{_CONFIG}->{db_name},
                                         host   => $self->{_CONFIG}->{db_host},
                                         user   => $self->{_CONFIG}->{db_user},
                                         pass   => $self->{_CONFIG}->{db_pass}
                                         );
    if (!$self->{_DB}) {
        $self->{Error} = "Could not make connection to database\n";
        return undef;
    }

    # Create user database (if different)
    if (defined $self->{_CONFIG}->{auth_db_name}
        && $self->{_CONFIG}->{auth_db_name} ne $self->{_CONFIG}->{db_name}) {
        $self->{_AUTH_DB} = 
            new Eidetic::Database(
                                  brand  => $self->{_CONFIG}->{auth_db_brand},
                                  name   => $self->{_CONFIG}->{auth_db_name},
                                  host   => $self->{_CONFIG}->{auth_db_host},
                                  user   => $self->{_CONFIG}->{auth_db_user},
                                  pass   => $self->{_CONFIG}->{auth_db_pass}
                                  );
        if (!$self->{_AUTH_DB}) {
            $self->{Error} = "Could not make connection to user database\n";
            return undef;
        }

    } else {
        # Our user/project/etc. tables are in the same database so just copy
        $self->{_AUTH_DB} = $self->{_DB};
    }

    # Force testing/setting of the old non-config fields
    $args{'fields'}->{'applicationTemplates'} = undef
        if (!defined($args{'fields'}->{'applicationTemplates'}));
    $args{'fields'}->{'globalHtmlRoot'} = undef
        if (!defined($args{'fields'}->{'globalHtmlRoot'}));
    $args{'fields'}->{'globalTemplates'} = undef
        if (!defined($args{'fields'}->{'globalTemplates'}));
    $args{'fields'}->{'globalHeader'} = "header.tmpl"
        if (!defined($args{'fields'}->{'globalHeader'}));
    $args{'fields'}->{'globalFooter'} = "footer.tmpl"
        if (!defined($args{'fields'}->{'globalFooter'}));
    $args{'fields'}->{'proj_update_email'} = undef
        if (!defined($args{'fields'}->{'proj_update_email'}));
    $args{'fields'}->{'project_notification_emails'} = undef
        if (!defined($args{'fields'}->{'project_notification_emails'}));
    $args{'fields'}->{'project_notification_emails_jp'} = undef
        if (!defined($args{'fields'}->{'project_notification_emails_jp'}));
    $args{'fields'}->{'webmaster'} = undef
        if (!defined($args{'fields'}->{'webmaster'}));
    $args{'fields'}->{'max_folder_depth'} = 12
        if (!defined($args{'fields'}->{'max_folder_depth'}));
    $args{'fields'}->{'default_folder'} = 1
        if (!defined($args{'fields'}->{'default_folder'}));
    $args{'fields'}->{'limit_view_user'} = 1
        if (!defined($args{'fields'}->{'limit_view_user'}));

    $self->{Error}="";
    # Validate fields from config file with our defaults
    foreach my $f (keys %{ $args{'fields'} }) {
        # if we didn't load it from the config file...
        if (!defined($self->{_CONFIG}->{$f})) {
            if (!defined($args{'fields'}->{$f})) {
                # there is no default...
                $self->{Error} .= "Must define '$f' in '$confname'\n";
            }
            else {
                # Default value gets set
                $self->{_CONFIG}->{$f}=$args{'fields'}->{$f};
            }
        }
    }
    return undef if ($self->{Error} ne "");


    # Configure our TT2 processor
    my $tt2_config = $args{'tt2_config'};
    foreach my $optname (keys %{$self->{_CONFIG}}) {
        if ($optname =~ /^TT2_(.*)$/) {
            my $opt=$1;
            if (!defined($tt2_config) || !defined($tt2_config->{$opt})) {
                $tt2_config->{$opt} = $self->{_CONFIG}->{"TT2_$opt"};
            }
        }
    }
    $self->{_CONFIG}->{'tt2_config'} = $tt2_config;
    $self->{_template} = Template->new($tt2_config);
    if (!defined($self->{_template})) {
        $self->{Error}="Template Toolkit did not initialize\n";
        return undef;
    }


    # == End of configuration section =============

    my $my_url     = $self->{_cgi}->url;
    my $auth_user  = $self->{_cgi}->remote_user;

    my $modulename = $self->{_cgi}->param('modulename')   || $self->{_CONFIG}->{default_modulename};
    my $command    = $self->{_cgi}->param('command')      || $self->{_CONFIG}->{default_command} || 'userpage';
    my $query      = $self->{_cgi}->param('query')        || '';
    my $filter     = $self->{_cgi}->param('filter')       || '';
    my $sort       = $self->{_cgi}->param('sort')         || '';
    my $maxitems   = $self->{_cgi}->param('maxitems')     || $self->{_CONFIG}->{default_maxitems} || 25;
    my $start      = $self->{_cgi}->param('start')        || 0;
    my $on         = $self->{_cgi}->param('on')           || '';
    my $hilite     = $self->{_cgi}->param('hilite')       || '';
    my $language   = lc($self->{_cgi}->param('language')) || $self->{_CONFIG}->{default_language} || 'en';

    $self->{_ARGS}->{my_url}=$my_url;
    $self->{_ARGS}->{auth_user}=$auth_user;

    $self->{_ARGS}->{modulename}=$modulename;
    $self->{_ARGS}->{command}=$command;
    $self->{_ARGS}->{query}=$query;
    $self->{_ARGS}->{filter}=$filter;
    $self->{_ARGS}->{sort}=$sort;
    $self->{_ARGS}->{maxitems}=$maxitems;
    $self->{_ARGS}->{start}=$start;
    $self->{_ARGS}->{on}=$on;
    $self->{_ARGS}->{hilite}=$hilite;
    $self->{_ARGS}->{language}=$language;

    return 1;
}


=head2 getConfig

Returns the loaded configuration hash.

=cut

sub getConfig {
    my $self = shift;
    $self->trace("getConfig") if ($self->{_TRACE});

    return $self->{_CONFIG};
}

=head2 getDatabase

Returns Eidetic::Database connection.

=cut

sub getDatabase
{
    my ($self) = @_;
    return $self->{_DB};
}

=head2 useObject(OBJECT)

If external code callbacks are going to be used by an instance of Eidetic,
this is how you pass an external object into Eidetic so that the data will
be available to the functions.  Internally, this is available as
$self->{_subobj}

=cut

sub useObject {
    my ($self, $obj) = @_;
    $self->trace("useObject") if ($self->{_TRACE});

    $self->{_subobj}=$obj;
}


=head2 registerModule(ARGS)

Registers a module name, table, commands, and queries into Eidetic.

 name        => 'users',
 table       => 'user',  
 commands    => [ 'edit', 'reedit', 'store', 'thanks', ],
 queries     => {
                    query_name => [ 'table', 'filter', 'sorting' ],
                },
 queries_from_database => 0,

If queries_from_database is non-zero, registerModule will load the
queries from the 'query' database table.

=cut

sub registerModule {
    my ($self, %args) = @_;
    $self->trace("registerModule") if ($self->{_TRACE});

    my $modulename = $args{'name'};
    if (!defined($modulename)) {
        $self->{Error}="No module name provided to registerModule()\n";
        return undef;
    }
    warn "Module $modulename\n" if ($self->{_DEBUG}>3);

    my $tablename  = $args{'table'};
    if (!defined($tablename)) {
        $self->{Error}="No table name provided to registerModule()\n";
        return undef;
    }

    my $cmdmap;
    $self->{Error}="";
    foreach my $cmd (@{$args{'commands'}}) {
    	if (!$self->can("do_$cmd")) {
            $self->{Error}.="Undefined command $cmd requested in registerModule()\n";
        }
        else {
            $cmdmap->{$cmd} = 1;
        }
    }
    return undef if ($self->{Error} ne "");

    # TODO: remove old-style?  is "database" used at all?
    # 
    # "database" was intended to allow multiple databases to be
    # referenced by an application (such as allowing intercommuncation
    # with RT) but is not used and probably requires rethinking anyway
    # before it could be implemented.
    $self->{_modules}->{$modulename} = {
        tables    => ["eidetic:$tablename"],
        database  => ['eidetic'],
        commands  => $cmdmap
    };

   if ($self->{_DEBUG}>3) {
        foreach my $info (sort keys %args) {
            warn "  have: $info\n";
        }
    }

    # Register any queries
    if (defined($args{'queries'})) {
        warn "   got queries\n" if ($self->{_DEBUG}>3);
        $self->registerQuery(modulename => $modulename,
                             queries => $args{'queries'});
    }

    # If the module is registered as allowing queries to be loaded from
    # the query table, we do that at this point.
    if (defined($args{'queries_from_database'})
        && $args{'queries_from_database'} != 0
        ) {
        warn "   loading queries from database\n" if ($self->{_DEBUG}>3);
        my $DB = $self->{_DB} || die "Cannot load queries:  Database not defined.\n";
        my $sth = $DB->selectRecords('query', "Module=$modulename", '', 0, 1);
        if ($sth != 0) {
            # Slurp in all matching rows and register them in memory
            my $queries = $sth->fetchall_arrayref;

            $self->registerQuery(modulename => $modulename,
                                 queries => $queries);
        }
    }

    return 1;
}


=head2 dispatch(ARGS)

This function is only called from inside the Eidetic "main" function.  It
sends a command out for execution.

 modulename => which module identity,
 command    => command to execute,
 on         => selection information,
 user       => user data hash,
 input_data => form/input data hash,

=cut
sub dispatch {

    my ($self,%args) = @_;
    $self->trace("dispatch") if ($self->{_TRACE});

    # Clean this up to do a nicer job of sending arguments
    my ($DB, $cgi, $modulename, $command, $on, $user, $input_data);

    $DB         = $self->{_DB};
    $cgi        = $self->{_cgi};

    $modulename = $args{'modulename'};
    $command    = $args{'command'};
    $on         = $args{'on'};
    $user       = $args{'user'};
    $input_data = $args{'input_data'};

    my ($content, %form);
    my @emails  = ();
    die trace_tree()."No tables defined for $modulename\n" 
    	unless ($self->{_modules}->{$modulename}->{tables});
    my $commands = $self->{_modules}->{$modulename}->{commands};

    my @tables = @{$self->{_modules}->{$modulename}->{tables}};
    my ($database,$table) = split /\:/, $tables[0];
    die trace_tree()."Invalid modulename; can't find matching database '$database'" 
        unless $database;
    die trace_tree()."Invalid modulename; can't find matching table '$table'" 
        unless $table;

    my %defaults;
    if ($self->{_modules}->{$modulename}->{defaults}) {
        %defaults = $self->{_modules}->{$modulename}->{defaults};
    }
    $user->{user_id} = $user->{$self->{_idField}};

    die trace_tree()."Command $command is not valid for '$modulename'\n"
    	unless (exists $commands->{$command});

    my $do_command;
    # First (main) command
    $do_command="do_$command";
    my ($newCommand, $newQuery, $noticeMsg, $newOn, $newModulename) =
    	$self->$do_command(
                                modulename  => $modulename,
                                command     => $command,
                                table       => $table, 
                                on          => $on,
                                data        => $input_data,
                                emails      => \@emails,
                                nextCommand => $command,
                                nextQuery   => $self->{_ARGS}->{query},
                                user        => $user);

    die trace_tree()."Command $newCommand is not valid for '$modulename'\n"
    	unless (exists $commands->{$newCommand});

    # Second follow-on command
    $newOn ||= $on;
    $newModulename ||= $modulename;
    if ($newCommand ne $command || $newQuery ne $self->{_ARGS}->{query}) {
        # Prevent trying to display the just-deleted item
        if ($command eq 'remove') { $newOn = ''; }
        
        my @tables = @{$self->{_modules}->{$newModulename}->{tables}};
        my ($database,$table) = split /\:/, $tables[0];
        die trace_tree()."Invalid modulename; can't find matching database '$database'" 
            unless $database;
        die trace_tree()."Invalid modulename; can't find matching table '$table'" 
            unless $table;

        $do_command="do_$newCommand";
        $self->$do_command(
                                   modulename => $newModulename,
                                   command    => $newCommand,
                                   table      => $table, 
                                   on         => $newOn,
                                   data       => $input_data,
                                   emails     => \@emails, 
                                   user       => $user,
                                   noticeMsg  => $noticeMsg);
    }
    return;
}


=head2 sendto_screen(VARS, TO_PROCESS)

Takes the "VARS" hashref, and uses it to process the page presentation
templates defined during "configure".  Any named template contents are
passed by the "TO_PROCESS" hash are processed first, and each result is
added to the "VARS" hash, by the name of the template in "TO_PROCESS".

 Example input:
  VARS
   title => "This is my title",
   content => "Body of work",
   etc...
 
  TO_PROCESS
   template1_name => raw template1 string,
   template2_name => raw template2 string,
   etc...

 Results in...
  VARS
   title => "This is my title",
   content => "Body of work",
   template1_name => rendered template1 results,
   template2_name => rendered template2 results,
   etc...
  
This function should only be called at the end of Eidetic command functions.

=cut

sub sendto_screen {
    my ($self, $var, $to_process) = @_;
    $self->trace("sendto_screen") if ($self->{_TRACE});

    die "Template Toolkit object not defined" 
	unless (defined $self->{_template});

    # $left_column is a placeholder for now, until the link box editor
    # is in place.  Perhaps the column text could be autogenerated in
    # this routine?
    
    my $tmplText = loadFile(catfile($self->{_CONFIG}->{applicationTemplates},
                                    "main.tmpl"));
    my $header   = loadFile(catfile($self->{_CONFIG}->{globalTemplates},
                                    $self->{_CONFIG}->{globalHeader}));
    my $footer   = loadFile(catfile($self->{_CONFIG}->{globalTemplates},
                                    $self->{_CONFIG}->{globalFooter}));

    $var->{'my_url'}        = $self->{_cgi}->url;
    $var->{'hostname'}      = hostname();
    $var->{'permissions'}   = defined($self->{_user}) && defined($self->{_user}->{Permissions}) ? $self->{_user}->{Permissions} : '';
    $var->{images_url}      = $self->{_CONFIG}->{images_url};

    # Set up defaults
    $var->{'to'}            ||= '';
    $var->{'left_column'}   ||= '';
    $var->{'right_column'}  ||= '';

    # TODO: don't prune out any TT2-style header/footer includes 
    my $content = $var->{'content'};
    $content =~ s/\[\%\s*INCLUDE "header.inc"[^%]*\%\]//;
    $content =~ s/\[\%\s*INCLUDE "footer.inc"[^%]*\%\]//;

    # Clear header, footer, and content
    $var->{'header'}        = '';
    $var->{'footer'}        = ''; 
    $var->{'content'}       = '';

    # Globals
    $var->{'GLOBALS'}       = $self->{_GLOBALS};
    $var->{'CONFIG'}        = $self->{_CONFIG};

    foreach my $template_name (sort keys %$to_process) {
        $var->{$template_name} = "";
        $self->{_template}->process(\$to_process->{$template_name},
                                    $var, \$var->{$template_name})
            || die trace_tree()."Template process failed on $template_name:  ".
            $self->{_template}->error()."\n";
    }

    # TODO:  I'll bet there's a way to process all these in one go...
    # Run content through template-toolkit
    $self->{_template}->process(\$content, $var, \$var->{content})
    	|| die trace_tree()."Template process failed on content:  ".
                            $self->{_template}->error()."\n";
    # Run header through template-toolkit
    $self->{_template}->process(\$header, $var, \$var->{header})
    	|| die trace_tree()."Template process failed on header:  ".
                            $self->{_template}->error()."\n";
    # Run footer through template-toolkit
    $self->{_template}->process(\$footer, $var, \$var->{footer})
    	|| die trace_tree()."Template process failed on footer:  ".
                            $self->{_template}->error()."\n";
    # Process the page and print result to stdout
    $self->{_template}->process(\$tmplText, $var);
}


=head2 authorizeUser(UID)

Return true if user $uid = HTTP-authorized user, false if not.

=cut

sub authorizeUser {
    my ($self, $uid) = @_;
    $self->trace("authorizeUser") if ($self->{_TRACE});
    my $DB=$self->{_USER_DB};

    if (!defined($uid) || $uid !~ /^\d+$/) {
        return 0;
    }
    if (! $DB->isInDatabase('user', "$self->{_idField}='$uid'")) {
        return 0;
    }

    my $sth = $DB->selectRecords('user', "$self->{_idField}=$uid", '', 0, 1);
    if ($sth == 0) {
        return 0;
    }
    my $user = $sth->fetchrow_hashref;

    if ($user->{descriptor} ne $self->{_ARGS}->{auth_user}) {
        return 0;
    }

    return 1;
}


=head2 getFieldText(DB, DATA, SOURCE, FIELD)

Gets the value for a given field from a given source.  This is what
parses the Eidetic-style fields to give autoinclusion of database
parameters.

=cut

sub getFieldText {
    my $self = shift;
    $self->trace("getFieldText") if ($self->{_TRACE});

    my $DB     = $self->{_DB} || die trace_tree()."getFieldText:  Undefined db handle";
    my $data   = shift || die trace_tree()."getFieldText:  Undefined data record";
    my $source = shift || die trace_tree()."getFieldText:  Undefined source table";
    my $field  = shift || die trace_tree()."getFieldText:  Undefined field";
    
    my $table = $source;

    # These two tables come from the _AUTH_DB database (which might be the
    # same as _DB)
    if ($table eq 'user' || $table eq 'project') {
        $DB = $self->{_AUTH_DB};
    }

    if ($source =~ /^\w+\_$self->{_idField}\_(\w+)$/) {
        $table = $1;
    } elsif ($source =~ /^\w+\_by$/
             || $source =~ /^\w+\_to$/) {
        $table = 'user';
    }
    die trace_tree()."Table undefined" unless $table;
    my $sourceId = 0;
    if (exists $data->{"$source\_$self->{_idField}"}) {
        $sourceId = $data->{"$source\_$self->{_idField}"};
	warn "Using data -> {$source\_$self->{_idField}}\n" if $self->{_DEBUG}>3;
    } elsif (exists $data->{$source}) {
        $sourceId = $data->{$source};
	warn "Using data -> {$source}\n" if $self->{_DEBUG}>3;
    } else {
	warn "sourceId $sourceId doesn't exist in table $table\n" if $self->{_DEBUG}>0;
	return 'err';
    }
    die trace_tree()."idField not defined" unless $self->{_idField};

    # TODO:  This should return undef probably
    return '' unless $sourceId;

    my $sth = $DB->selectRecords(lc($table), "$self->{_idField}=$sourceId", '')
	|| die trace_tree()."getFieldText:  Could not select records:  $!";
    my $item = $sth->fetchrow_hashref;
    return $item->{$field};
}


=head2 do_userpage

Display the admin page for a user\'s account. More pseudo-TT stuff.
(Userid is $defaults->{user_id})

=cut

sub do_userpage {
    my ($self, %args) = @_;
    $self->trace("do_userpage") if ($self->{_TRACE});

    my $modulename=$args{'modulename'};
    my $command=$args{'command'};
    my $table=$args{'table'};
    my $DB=$self->{_DB};
    my $on=$args{'on'};
    my $dataHandle=$args{'data'};
    my $emails=$args{'emails'};
    my $nextCommand=$args{'nextCommand'};
    my $nextQuery=$args{'nextQuery'};
    my $user=$args{'user'};
    my $noticeMsg=$args{'noticeMsg'};

    my $maxitems = 10;

    my ($content, $right_column);
    my $user_id = $user->{user_id} ||
        die trace_tree()."User ID number not found";

    $self->{_ARGS}->{filter} =~ s/\[\%\s*user_id\s*\%\]/$user_id/g;

    # Load up the right hand column of commands
    my $showsection = 1;
    my $title = '';
    open(RIGHTCOLUMN, "<$self->{_CONFIG}->{applicationTemplates}/$modulename/display.right_column.tmpl");
    while (<RIGHTCOLUMN>) {
        if ($self->{_ARGS}->{auth_user}) {
            s/\[\%\s*Username\s*\%\]/$self->{_ARGS}->{auth_user}/g;
            s/\[\%\s*$self->{_idField}\s*\%\]/$user_id/g;
        } else {
            s/\[\%\s*Username\s*\%\]/Anonymous/g;
            s/\[\%\s*$self->{_idField}\s*\%\]/NONE/g;
        }
        # User commands
        while (/\[\%\s*command\s*\(\s*\"(\w+)\"\,\s*\"([\w\&\=\-]+)\"\,\s*\"([\w\s]+)\"\s*\)\s*\%\]/) {
            my $action = '';
            if ($showsection == 1) {
                my $cmd = $self->cgiCmd($1, $2);
                my $item = qq|<A HREF="$cmd">$3</A><BR>|;
                $action = loadFile("$self->{_CONFIG}->{applicationTemplates}/box_row.tmpl");
                $action =~ s/\[\%\s*item\s*\%\]/$item/;
            }
            s/\[\%\s*command\s*\(\s*\"(\w+)\"\,\s*\"([\w\&\=\-]+)\"\,\s*\"([\w\s]+)\"\s*\)\s*\%\]/$action/;
        }
        while (/\[\%\s*section\s*\(\s*\"(\w+)\"\,\s*\"([\s\w]+)\"\s*\)\s*\%\]/) {
            my $sec = '';
            $title = $2;
            # Check to see if user has permissions for this area
            if (exists $user->{Permissions}->{$1}
                && $user->{Permissions}->{$1} eq 'Y') {
                $showsection = 1;
                $sec = loadFile("$self->{_CONFIG}->{applicationTemplates}/box_start.tmpl");
                $sec =~ s/\[\%\s*box_title\s*\%\]/$title/g;
            } else {
                $showsection = 0;
            }
            s/\[\%\s*section\s*\(\s*\"(\w+)\"\,\s*\"([\s\w]+)\"\s*\)\s*\%\]/$sec/;
        }
        while (/\[\%\s*altsection\s*\%\]/) {
            my $alt = "";

            if ($showsection == 0) {
                $showsection = 1;
                $alt = loadFile("$self->{_CONFIG}->{applicationTemplates}/box_start.tmpl");
                $alt =~ s/\[\%\s*box_title\s*\%\]/$title/;
            } else {
                $showsection = -1;
            }
            # Check to see if user has permissions for this area
            s/\[\%\s*altsection\s*\%\]/$alt/;
        }
        while (/\[\%\s*endsection\s*\%\]/) {
            my $end = "";
            if ($showsection != 0) {
                $end = loadFile("$self->{_CONFIG}->{applicationTemplates}/box_end.tmpl");
            } else {
                $showsection = 1;
            }

            # Check to see if user has permissions for this area
            s/\[\%\s*endsection\s*\%\]/$end/;
        }
        $right_column .= $_;

    }
    close(RIGHTCOLUMN);

    $content = $self->do_single(modulename=>$modulename,
                                command=>'display',
                                table=>$table,
                                on=>$on,
                                user=>$user,
                                noticeMsg=>$noticeMsg,
                                returncontent=>1);

    $noticeMsg="" unless (defined($noticeMsg));
    $self->sendto_screen({
                         title=>ucfirst($modulename),
                         Notice => $noticeMsg,
                         content=>$content,
                         },
                         { right_column => $right_column, });
    return ($nextCommand, $nextQuery, $noticeMsg);
}

=head2 do_redirect

Generates a redirect to $dataHandle->{path} & {file}.

=cut

sub do_redirect {
    my ($self, %args) = @_;
    $self->trace("do_redirect") if ($self->{_TRACE});

    my $path = $args{'data'}->{path} || "";
    my $file = $args{'data'}->{file} || "";

    my $uri = $path . $file;
    print $self->{_cgi}->redirect($uri);

    return ($args{'nextCommand'}, $args{'nextQuery'}, $args{'noticeMsg'});
}

=head2 do_search

Loads a template from APPLICATION/MODULE/search.tmpl to generate an
input form.  On matched records, "display" command is used in the "many"
context to show results.

When the "search" command is run all the web form fields that start with
"d_" will be processed as possible fields to search against.

The module's current table is searched for any matching fields.  An empty
field is treated as "match anything", and "*" characters are treated as
wildcards.

For example, if the module's table has a field named "Notes", the
template would contain "d_Notes", and all resulting records matching the
input would be displayed through the "display" command.

Any field with Eidetic's "id field" ("uid" by default) in it, will be
processed in more detail.  Everything up to "uid" is processed as the name
of a field to match against.  The format of these fields can be:

 TABLE_uid
    Search table "TABLE" on the field "descriptor" and report any matched
    records that are listed in the module table's "TABLE_uid" field.

 TABLE_uid.FIELD
    Search table "TABLE" on the field "FIELD" and report any matched
    records that are listed in any of module table's "TABLE_uid" field.

 field_name_uid_TABLE
    Search table "TABLE" on the field "descriptor" and report any matched
    records that are listed in any of module table's "field_name_uid" field.

 field_name_uid_TABLE.FIELD
    Search table "TABLE" on the field "FIELD" and report any matched
    records that are listed in any of module table's "field_name_uid" field.

=cut

sub do_search {
    my ($self, %args) = @_;
    $self->trace("do_search") if ($self->{_TRACE});

    my $modulename=$args{'modulename'};
    my $command=$args{'command'};
    my $table=$args{'table'};
    my $DB=$self->{_DB};
    my $on=$args{'on'};
    my $dataHandle=$args{'data'};
    my $emails=$args{'emails'};
    my $nextCommand=$args{'nextCommand'};
    my $nextQuery=$args{'nextQuery'};
    my $user=$args{'user'};
    my $noticeMsg=$args{'noticeMsg'};

    if (!defined($dataHandle) || !%$dataHandle) {
        my $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/search.tmpl");

	    $self->sendto_screen({title=> "Advanced Search",
                                  content=>$content,
                                  Notice=>$noticeMsg});

        return ($nextCommand, $nextQuery, $noticeMsg);
    }
    while (my ($key, $value) = each %$dataHandle) {
        # TODO:  Add ability to handle non-strings
        # TODO:  Add capability to specify GT, LT, EQ, CN, etc.
        if ($value ne '*' && $value ne '') {
            if ($self->{_ARGS}->{filter} ne '') {
                $self->{_ARGS}->{filter} .= " AND ";
            }
            # Don't use wild cards unless we use "*"
            $value = quotemeta($value);
            $value =~ s/\\\*/\%/g;

            # Search field format can be:
            # TABLE_uid             (FIELD defaults to "descriptor")
            # name_uid_TABLE        (FIELD defaults to "descriptor")
            # name_uid_TABLE.FIELD
            if ($key =~ /^((\w+)\_$self->{_idField}(\_(\w+))?)(\.(\w+))?$/) {
                my $table_key = $1;
                my $source = $2;
                my $field = "descriptor";
                
                # Which part contains the source table information?
                if ($3 ne "") {
                    $source=$4;
                }
                # Which field do we need to search on?
                if ($5 ne "") {
                    $field = $6;
                }

                my $sth = $DB->selectRecords(lc($source),
                                        "$field LIKE '$value'", '');
                my @ids = ();
                while (my $item = $sth->fetchrow_hashref) {
                    my $sourceId = $item->{$self->{_idField}};
                    if ($sourceId) {
                        push @ids, $sourceId;
                    }
                }
                if (scalar(@ids) >= 1) {
                    $self->{_ARGS}->{filter} .=
                                    "$table_key IN (".join(',',@ids).")";
                } else {
                    die trace_tree().
                        "The search item '$value' matches no known '$source'";
                }
            } else {
                $self->{_ARGS}->{filter} .= "$key LIKE '$value'";
            }
        }
    }
    $self->do_many(modulename=>$modulename,
                   command=>$command,
                   table=>$table,
                   on=>$on,
                   user=>$user);

    return ($nextCommand, $self->{_ARGS}->{query}, $noticeMsg);
}

=head2 do_display

Uses do_single, do_many, or do_tree to display any records matching 
the "on" and "query" arguments.

The first template found for a given context is used to display the results.
The templates are:

  on=number ("single" context via "do_single")
    any COMMAND
      APPLICATION/MODULE/COMMAND.single.tmpl

    
  on=number-number,number,... ("many" context via "do_many")
    Multiple templates are used for display:
      page   (which contains [% header %] and [% row %]
      header
      row

    For the "page" template:

      COMMAND=edit
        APPLICATION/MODULE/edit.list.form.tmpl
      COMMAND=search
        APPLICATION/MODULE/display.all.page.tmpl
      COMMAND=anything else
        APPLICATION/MODULE/COMMAND.QUERY.page.tmpl

    For the "header" template:
      COMMAND=any
        APPLICATION/MODULE/COMMAND.QUERY.header.tmpl
        APPLICATION/MODULE/COMMAND.header.tmpl
        APPLICATION/MODULE/display.all.header.tmpl

    For the "row" template:
      COMMAND=any
        APPLICATION/MODULE/COMMAND.QUERY.row.tmpl
        APPLICATION/MODULE/COMMAND.row.tmpl
        APPLICATION/MODULE/display.all.row.tmpl
      

  on=number[number] ("tree" context via "do_tree")
    Uses object-defined templates for display.  See "do_tree"


=cut

sub do_display {
    my ($self, %args) = @_;
    $self->trace("do_display") if ($self->{_TRACE});

    my $modulename=$args{'modulename'};
    my $command=$args{'command'};
    my $table=$args{'table'};
    my $DB=$self->{_DB};
    my $on=$args{'on'};
    my $dataHandle=$args{'data'};
    my $emails=$args{'emails'};
    my $nextCommand=$args{'nextCommand'};
    my $nextQuery=$args{'nextQuery'};
    my $user=$args{'user'};
    my $noticeMsg=$args{'noticeMsg'};

    die trace_tree()."Error - table name not given for $command $modulename (on '$on')\n" 
        unless $table;

    if ($modulename eq 'users' && $user->{$self->{_idField}} ne $on) {
        if ($user->{Permissions}->{admin} ne 'Y'
            && $self->{_CONFIG}->{limit_view_user}) {
            die trace_tree()."Cannot access record for $on\n";
        }
    }

    if ($self->{_ARGS}->{query} =~ /^[A-Za-z]\w+$/) {
        ($table, $self->{_ARGS}->{filter}, $self->{_ARGS}->{sort}) = $self->getQuery($modulename, $self->{_ARGS}->{query});
        if ($on eq '') {
            $on = "0-";
        }
        $self->{_ARGS}->{filter} =~ s/\[\%\s*user_id\s*\%\]/$user->{user_id}/g;
        die trace_tree()."I don't know what the query '$self->{_ARGS}->{query}' means<BR>\n" 
            unless ($table);
    } elsif ($on eq '') {
        $on = '0-';
    }
    if ($self->{_ARGS}->{query} eq '') {
        $nextQuery = 'all';
        $self->{_ARGS}->{query} = 'all';
    }

    my $parent_field = 'Parent_uid';
    
    if ($on =~ /^(\d+)\[(\d+)\]$/) {
        $self->trace("Calling do_tree for $1 of depth $2\n") if ($self->{_TRACE});

        $self->do_tree(modulename=>$modulename,
                       command=>$command,
                       table=>$table,
                       parent_field=>$parent_field,
                       parent_id=>$1,
                       depth=>$2,
                       user=>$user,
                       noticeMsg=>$noticeMsg);

    } elsif ($on =~ /^\d+$/) {
        $self->do_single(modulename=>$modulename,
                         command=>$command,
                         table=>$table,
                         on=>$on,
                         user=>$user,
                         noticeMsg=>$noticeMsg);
    } elsif ($on =~ /^[\-\d]\d*[\!\,\-][\,\-\d]*$/) {
        $self->do_many(modulename=>$modulename,
                       command=>$command,
                       table=>$table,
                       on=>$on,
                       user=>$user,
                       noticeMsg=>$noticeMsg);
    } else {
    	warn "Unknown on=$on specified\n" if $self->{_DEBUG};
    }
    return ($nextCommand, $nextQuery, $noticeMsg);
}

=head2 do_edit

Displays, via HTML, the edit form for $modulename and $query. Uses do_single and
do_many.

The data hash is used for providing default values for the edit form.
These are provided via the URL as keys prepended with "d_".

=cut

sub do_edit {
    my ($self, %args) = @_;
    $self->trace("do_edit") if ($self->{_TRACE});

    my $modulename=$args{'modulename'};
    my $command=$args{'command'};
    my $table=$args{'table'};
    my $DB=$self->{_DB};
    my $on=$args{'on'};
    my $data=$args{'data'};
    my $emails=$args{'emails'};
    my $nextCommand=$args{'nextCommand'};
    my $nextQuery=$args{'nextQuery'};
    my $user=$args{'user'};
    my $noticeMsg=$args{'noticeMsg'};

    my $template_file = "";

    if ($self->{_ARGS}->{query} ne '') {
        ($table, $self->{_ARGS}->{filter}, $self->{_ARGS}->{sort}) 
            = $self->getQuery($modulename, $self->{_ARGS}->{query});
    }

    my $content = '';
    my $hiddenInfo = '';

    # Check if the user is allowed to edit
    if ($on =~ /^\d+$/) {
    	# Special case for user editing
    	if ($modulename eq 'users' && $user->{$self->{_idField}} ne $on) {
    	    if ($user->{Permissions}->{admin} ne 'Y') {
                die trace_tree()."Cannot access record for $on\n";
    	    }
    	}
    } 

    # Get metadata about the form from the database
    my %form = (
                fields =>       $DB->getTableInfo($table, 'INPUT_FIELDS'),
                descriptions => $DB->getTableInfo($table, 'DESCS'),
                types =>        $DB->getTableInfo($table, 'TYPES'),
                notes =>        $DB->getTableInfo($table, 'EXTRA'),
                defaults =>     $DB->getTableInfo($table, 'DEFAULTS')
                );

    # Load up the data record, if appropriate
    if ($on =~ /^\d+$/) {
        die trace_tree()."Couldn't find $on\n" unless $DB->isInDatabase($table, "$self->{_idField}=$on");

        my $sth = $DB->selectRecords($table, "$self->{_idField}=$on")
             || die trace_tree()."Couldn't process $on!\n";
        $data = $sth->fetchrow_hashref;

        if ($modulename eq 'projects' &&
            $user->{user_id} != $data->{Lead_uid_user}
            && $user->{Permissions}->{projects_admin} ne 'Y') {
            die trace_tree()."You cannot edit this project.  User is: " 
                . ($user->{user_id}) 
                . " and project owner is " . $data->{Lead_uid_user};
        }

        $form{values} = $data;


        # Single item...
        $template_file = catfile($self->{_CONFIG}->{applicationTemplates},
                                 $modulename);
        if (-e catfile($template_file, "edit.single.tmpl")) {
            $template_file = catfile($template_file, "edit.single.tmpl");
        } elsif (-e catfile($template_file, "edit.$table.form.tmpl")) {
            $template_file = catfile($template_file, "edit.$table.form.tmpl");
        } else {
            $template_file = "";
        }
    } else {
        $form{values} = $form{defaults};
        # Allow overrides from the user
        foreach my $key (keys %$data) {
            if (defined ($form{types}->{$key}) ) {
                if ($data->{$key} !~ /^\d+$/ && $key =~ /^(\w+)\_$self->{_idField}/) {
                    # Convert descriptor to id
                    $data->{$key} = $DB->getId(lc($1), $data->{$key});
                }
                $form{values}->{$key} = $data->{$key};
            } elsif (defined $form{types}->{"$key\_$self->{_idField}"}) {
                # Convert descriptor to id
                $data->{"$key\_$self->{_idField}"} = $DB->getId(lc($key), $data->{$key});
            }
        }

        # The template names to try, in order
        my @templates;
        if (!defined($on) || $on eq "") {
            # Default template for "new" entries
            push(@templates,catfile($self->{_CONFIG}->{applicationTemplates},
                            $modulename,
                            "edit.$table.new.form.tmpl"));
        }
        push(@templates, catfile($self->{_CONFIG}->{applicationTemplates},
                            $modulename,
                            "edit.$table.$self->{_ARGS}->{query}.form.tmpl"));
        push(@templates, catfile($self->{_CONFIG}->{applicationTemplates},
                            $modulename,
                            "edit.$table.form.tmpl"));

        # Find the first template that exists
        foreach my $possible (@templates) {
            if (-r $possible) {
                $template_file=$possible;
                last;
            }
        }
    }    

    # If we can load the template, use it, else generate a default page
    if ($template_file ne "" && -e $template_file) {
        $content = loadFile($template_file);
    } else {
        $content = $self->genericHtmlForm($modulename, 'store', %form);
    }

    my $vars = {
	%{$self->{_storageError}},
	Hidden_info      => $hiddenInfo,
	Notice           => $noticeMsg,
	user_id          => $user->{user_id},
	Real_email       => $user->{Real_email} || '',
	Username         => $user->{descriptor} || '',
        %$data
    };

    if ($on eq '' || $on =~ /^\d+$/) {
    	$content = $self->substituteValues($content, $form{values}, '');
    	$content = $self->processPage($content, $data);
    	$self->sendto_screen({
            title=>ucfirst("$modulename:  Submit $table"),
            content=>$content,
            %$vars
            });
        
    } elsif ($on =~ /^[\-\d]\d*[\!\,\-][\,\-\d]*$/) {
        $self->do_many(modulename=>$modulename,
                       command=>$command,
                       table=>$table,
                       on=>$on,
                       user=>$user);
    } else {
        # User entered strange value for $on; need to ask for clarification...
    	die trace_tree()."Unknown on parameter '$on'\n";
    }
    if ($self->{_ARGS}->{query} eq 'all') { 
        $self->{_ARGS}->{query}=''; 
        $nextQuery=$self->{_ARGS}->{query}; 
    }

    return ($nextCommand, $nextQuery, $noticeMsg);
}

=head2 validate(VALUE, FORMATTAG)

Validates that $value is a valid $formatTag; returns list with 2 empty elements
if so, list with 'reedit' and an error message otherwise.

=cut

sub validate {
    my ($self, $value, $formatTag) = @_;
    $self->trace("validate") if ($self->{_TRACE});
    my $errorMsg;

    # TODO:  Load the regexps from a database table

    if ($formatTag eq 'name' && $value !~ /^[\w\.\,\-\' ]{1,80}$/) {
        $errorMsg = "This must be between 1 and 80 characters in length and use only the characters 0-9, a-z, A-Z, and the following punctuation:  _.,-'";
    } elsif ($formatTag eq 'email' && $value !~ /^[\w\.\@\-]{5,255}$/) {
        $errorMsg = "This must be between 5 and 255 characters in length and use only the characters 0-9, a-z, A-Z, and the following punctuation: _.@-";
    } elsif ($formatTag eq 'username' && $value !~ /^[a-z0-9\_]{2,20}$/) {
        $errorMsg = "This must be between 2 and 20 characters in length and use only the characters 0-9, a-z, and _.";
    } else {
        return ('', '');
    }
    return ('reedit', $errorMsg);
}

=head2 do_store

Stores data into the given module\'s table. If $on is empty, adds a new record;
otherwise updates records given by $on. Special cases for users, and 
projects.

=cut
sub do_store {
    my ($self, %args) = @_;
    $self->trace("do_store") if ($self->{_TRACE});

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $DB          = $self->{_DB};
    my $on          = $args{'on'};
    my $dataHandle  = $args{'data'};
    my $emails      = $args{'emails'};
    my $nextCommand = $args{'nextCommand'};
    my $nextQuery   = $args{'nextQuery'};
    my $user        = $args{'user'};
    my $noticeMsg   = $args{'noticeMsg'};

    my $content = '';
    $nextCommand = 'display';
    my $nextOn = '';
    my $user_id = $user->{user_id} || 0;
    my $uploaded_filename = '';

    if ($on eq '') {
        my $message = '';
        my @next = ();
        %{$self->{_storageError}} = ();

        # Special code for various module
        if ($modulename ne 'documents' 
            && $DB->isInDatabase($table, 
                            "descriptor='".$dataHandle->{descriptor}."'")) {
            $content .= $dataHandle->{descriptor} . ' already exists.';
            $self->{_storageError}->{Descriptor_Error} =
                'This is already used, try another.';
            $next[0] = 'reedit';
        } elsif ($modulename eq 'users') {
            # Validate the user's info
            ($next[0], $self->{_storageError}->{Given_Name_Error}) =
                $self->validate($dataHandle->{Given_name}, 'name');
            ($next[1], $self->{_storageError}->{Surname_Error}) =
                $self->validate($dataHandle->{Surname}, 'name');
            ($next[2], $self->{_storageError}->{Real_Email_Error}) =
                $self->validate($dataHandle->{Real_email}, 'email');
            if ($next[2] eq '' && $DB->isInDatabase($table, "Real_email='"
                                               .$dataHandle->{Real_email}."'")) {
                $next[2] = 'reedit';
                $self->{_storageError}->{Real_Email_Error} =
                    "This email address is already registered.";
            }

            ($next[3], $self->{_storageError}->{Descriptor_Error}) =
                $self->validate($dataHandle->{descriptor}, 'username');
            $dataHandle->{Username} = $dataHandle->{descriptor};

            my $now = $DB->quickQuery("SELECT NOW()");

            $dataHandle->{Join_date} = $now;
        }

        my $new_file = '';
        if (join('', @next) ne '') {
            # Break here and require user to reedit
            $noticeMsg = "Please check your entry and re-edit...";
            return ('reedit', $self->{_ARGS}->{query}, $noticeMsg);
        } elsif ($modulename eq 'users') {
            # Generate a new password
            my $passwd = generateRandomPassword();

            # Prepare email msg for user
            $dataHandle->{Userpassword} = $passwd;
            $message = loadFile(catfile($self->{_CONFIG}->{applicationTemplates},'users','new.user.email.tmpl'));
            $message = $self->substituteValues($message, $dataHandle);
            # Encrypt password
            my $salt = sprintf("\$1\$%.8s\$", generateRandomPassword());
            $dataHandle->{Userpassword} = md5_hex($passwd);
            $dataHandle->{Unixpassword} = crypt($passwd, $salt);
            $dataHandle->{NTpassword}   = nthash($passwd);

            # We're going to forward them to a thankyou page next.
            $nextCommand = 'thanks';
        } elsif ($modulename eq 'documents') {
            if (! $dataHandle->{Filename} ) {
                $noticeMsg = 'Error in uploading file';
                $self->{_storageError}->{Filename_Error} =
                    "Filename not specified - cannot perform upload.";
                $nextCommand = 'reedit';
                return ($nextCommand, $self->{_ARGS}->{query}, $noticeMsg);
            } else {
                $dataHandle->{Submitter_uid_user} = $user_id;
                $dataHandle->{Author} = $user->{descriptor} || 'Anonymous';
                $dataHandle->{Status} ||= 'Released';
                $dataHandle->{Version} = 1;

                if (! $dataHandle->{Language_uid}) {
                    $dataHandle->{Language_uid} = 1001;
                }
 
		$uploaded_filename = $self->{_subobj}->receive_init($dataHandle);

                # Verify document filename is not already in the system
                if ($DB->isInDatabase('document', "Filename='".$dataHandle->{Filename}."' AND rsf=1")) {
                    die "Sorry, that title/filename ('$dataHandle->{Filename}') "
			."conflicts with documents already in "
                        ."the system.  Click the back button and try a different name.\n";
                }

            }
        }

        while (my ($key, $value) = each %$user) {
            if (! $dataHandle->{$key} 
                && $key ne 'descriptor'
                && $key ne $self->{_idField}
                && $key ne 'rsf') {
                $dataHandle->{$key} = $value;
            }
        }

	if ($modulename eq 'projects' 
# Ignore the Use_folders param for now - all projects get a folder.  
# We need to configure the project page to only show the 'Add folder'
# command when the project has requested to use folders.
#	    && defined $dataHandle->{Use_folders}
	    ) {
	    # Add a top-level folder for the project, as well
	    my %folder = $DB->validateAndStoreRecord('folders',
						     $user_id,
						     $self->{_modules}->{'folders'}->{tables},
						     %$dataHandle
						     );
	    $dataHandle->{Folder_uid} = $folder{$self->{_idField}};
	    $noticeMsg = 'Project created with a document folder';
	}

        # Validate the inputs and store the data
        my %data = $DB->validateAndStoreRecord($modulename, 
                            $user_id, 
                            $self->{_modules}->{$modulename}->{tables}, 
                            %$dataHandle);

        if ($modulename eq 'documents' && defined $data{Language_uid}) {
            my $sth = $DB->selectRecords('language', "uid=$data{Language_uid}", '');
            if ($sth != 0) {
		my $ln = $sth->fetchrow_hashref;
		if ($ln) {
		    $data{Language} = $ln->{descriptor};
		} else {
		    $data{Language} = $self->{_ARGS}->{language} || 'en';
		}
            }
        }

        # Perform file upload processing
        if (($modulename eq 'documents') 
            && $data{Filename} ne '') {

	    # Map id to uid
	    $data{id} = $data{$self->{_idField}};
	    while (my ($key, $value) = each %data) {
		next unless ($key);
		warn "$key = $value\n" if ($value && $self->{_DEBUG}>2);
	    }

            my $archive = $self->{_subobj}->create_document_from_files(\%data, $uploaded_filename);
            if (! $archive) {
                die trace_tree()."The file could not be entered into the repository.\n";
            }
            # Link the document
            $self->{_subobj}->create_document_link(
                                                   $data{id},
                                                   $data{Parent_resource_uid}
                                                   ) || carp "Could not attach file: $!\n";
            my $errorMsg = $self->{_subobj}->create_renderings($archive, 
					     $data{Filename}, 
					     undef,
					     "$self->{_CONFIG}->{docs_path}");
            die trace_tree()."Error processing document - $errorMsg\n" 
                if ($errorMsg);

	    if ($self->{_DEBUG}) {
		warn "Undefined data{Filename}"    unless ($data{Filename});
		warn "Undefined data{$self->{_idField}}"    unless ($data{$self->{_idField}});
		warn "Undefined data{Version}"     unless ($data{Version});
		warn "Undefined my_url"            unless ($self->{_ARGS}->{my_url});
		warn "Undefined webmaster"         unless ($self->{_CONFIG}->{webmaster});
	    }

            $noticeMsg = qq| <P><B>'$data{Filename}'</B> has been
                successfully received into the repository as document
                number $data{$self->{_idField}}, ver. $data{Version}.  You can
                access it directly using <A
                HREF="$self->{_CONFIG}->{docs_url}/$data{Filename}">this link</A>.

                <P>If you suspect any problems, please email 
                <A HREF="mailto:$self->{_CONFIG}->{webmaster}">$self->{_CONFIG}->{webmaster}</A>.</P>
                |;
        }

        if ($modulename eq 'users') {
            $self->sendto_email($dataHandle->{Real_email}, "User Account", $message);
        } elsif ($modulename eq 'projects') {
	    # TODO:  Need better hostname detection mechanism
            if (`hostname` !~ /^osdlab/) {
                $self->notifyStaff($modulename, %data);
            }
        }

        # Fall through to display routine down below.
        if ($noticeMsg eq '') {
            $noticeMsg = "Record $data{$self->{_idField}} was stored successfully.  ";
        }
        $nextOn = $data{$self->{_idField}};
	warn "Next will be $nextOn\n" if ($self->{_DEBUG}>2);
    } 
    my $file_uploaded;

    if ($on =~ /^\d+$/) {

        if ($modulename eq 'users' && $user->{$self->{_idField}} ne $on) {
            if ($user->{Permissions}->{admin} ne 'Y') {
                die trace_tree()."Cannot access record for $on - must have admin permissions\n";
            }
        }

        # Store edits to a particular record

        my $owner = 1;
#       $self->authorizeUser($self->{_ARGS}->{auth_user}, "uid='$owner'") ||
#           die trace_tree()."You do not have access to change this password.";

        if ($modulename eq 'documents') {
	    my $owner = $self->get_owner('document', $on, 'Submitter_uid_user');
	    if ($owner ne $user->{descriptor} && $user->{Permissions}->{admin} ne 'Y') {
		die "Only the owner of this document ($owner) can update it";
	    }

            # Get the current version number & verify this user can do editing
            my $old_filename = '';
            my $sth = $DB->selectRecords('document', "$self->{_idField}=$on", '');
            my $old_version;
            if ($sth != 0) {
                $old_version = $sth->fetchrow_hashref;
                $old_filename = $old_version->{Filename};
            }

            if (defined $dataHandle->{Filename} && $dataHandle->{Filename} ne '') {
                $dataHandle->{Version} = $old_version->{Version} + 1;
                $dataHandle->{Submitter_uid_user} ||= $user_id;

                $file_uploaded = $self->{_subobj}->receive_init($dataHandle);

                # Temporary solution to duplicate filenames; a better solution will be needed
                if ($dataHandle->{Filename} ne $old_filename
                    && $DB->isInDatabase('document', "Filename='".$dataHandle->{Filename}."' AND rsf=1")) {
                    die "Sorry, that title/filename (".($dataHandle->{Filename})
                        .") conflicts with documents already in the system. "
                        ." Click the back button and try a different name.\n";
                }
            } else {
                $dataHandle->{Filename} = $old_version->{Filename};
            }
        }
        # Validate the inputs and store the data
        my %data = $DB->updateRecord($modulename, 
				     $user_id, 
				     $self->{_modules}->{$modulename}->{tables}, 
				     %$dataHandle);

        if ($modulename eq 'documents' && defined $data{Language_uid}) {
            my $sth = $DB->selectRecords('language', "uid=$data{Language_uid}", '');
            if ($sth != 0) {
                $data{Language} = ($sth->fetchrow_hashref)->{Charset};
            }
        }

        # Perform file upload processing
        if ($modulename eq 'documents' && $data{Filename} && $file_uploaded) {

	    # Uninstall any old files from whereever they're installed
	    my $sth = $DB->selectRecords('document', "uid=$data{$self->{_idField}}", '');
	    if ($sth != 0) {
		my $fname = ($sth->fetchrow_hashref)->{Filename};
		$self->{_subobj}->remove_renderings($fname,
					   undef,
					   $self->{_CONFIG}->{docs_path}
					   );
	    }
	    
	    # map id to uid
	    $data{id} = $data{$self->{_idField}};
	    warn "Calling create_document_from_files\n" if ($self->{_DEBUG}>2);
	    while (my ($key, $value) = each %data) {
		next unless ($key);
		warn "$key = $value\n" if ($value && $self->{_DEBUG}>2);
	    }

            my $archive = $self->{_subobj}->create_document_from_files(\%data, $file_uploaded);
            if (! $archive) {
                die trace_tree()."The file could not be entered into the repository.\n";
            }
            my $errorMsg =
                $self->{_subobj}->create_renderings($archive, 
                                                    $data{Filename}, 
                                                    undef,
                                                    $self->{_CONFIG}->{docs_path});

            die trace_tree()."Error processing document - $errorMsg\n" unless (! $errorMsg);

	    $noticeMsg = qq|
		<P><B>'$data{Filename}'</B> has been successfully received 
		into the repository as document # $data{$self->{_idField}}, 
		ver. $data{Version}.  You can access it from the page of
		the project and/or folder owning it, 
		or directly using <A HREF="$self->{_CONFIG}->{docs_url}/$data{Filename}">this link</A>.
		
		<P>If you suspect any problems, please email 
		<A HREF="mailto:$self->{_CONFIG}->{webmaster}">$self->{_CONFIG}->{webmaster}</A>.</P>
		|;
        }

        if (!defined($noticeMsg) || $noticeMsg eq '') {
            $noticeMsg = "Record $data{$self->{_idField}} was updated successfully.  ";
        }
        $nextOn = $on;
    } 

    if ($on =~ /^\d+\-\d+$/ || $on =~ /^\d+\-$/) {
        if ($modulename eq 'users' && $user->{$self->{_idField}} ne $on) {
            if ($user->{Permissions}->{admin} ne 'Y') {
                die trace_tree()."Cannot store records for $on - must have admin permissions\n";
            }
        }

        # Storing a range of records
        my %records;

        # Split the data into an array of records indexed by id
        while (my ($key, $value) = each %$dataHandle) {
            if ($key =~ /$self->{_idField}\((\d+)\).(\w+)/) {
                $records{$1}->{$2} = $value;
            }
        }

        foreach my $rec_id (keys %records) {
            $records{$rec_id}->{$self->{_idField}} = $rec_id;
            my %rec = $DB->updateRecord($modulename, 
					$user_id, 
					$self->{_modules}->{$modulename}->{tables},
					%{ $records{$rec_id} });
        }
        if ($modulename eq 'projects') {
            $self->sendto_email($self->{_CONFIG}->{proj_update_email},
                         "Project Status Updates",
                         "The following project ID's have been updated:  "
                         . join(', ', keys %records)
                         );
        }

        # Fall through to display routine down below.
        if ($noticeMsg eq '') {
            $noticeMsg = "Records $on updated.";
        }
    }

    return ($nextCommand, $nextQuery, $noticeMsg, $nextOn);
}




=head2 makeWidget(DB, TYPE, DATA, TABLE, FIELD, DEFAULT, FIELD_LIST)

Makes various HTML widgets

=cut
sub makeWidget {
    my ($self, $type, $data, $table, $field, $default, $field_list, $query) = @_;
    $self->trace("makeWidget") if ($self->{_TRACE});
    my $DB=$self->{_DB};
    my $text = '';
#    my $dbname = "eidetic";
#    my $dbh = getDbHandle($dbname);
    my $sth;

    if (!$table || !$field) {
        # Bad info...
        return ($self->{_DEBUG}? 'Insufficient parameters' : '');
    } elsif ($DB->checkTableExistance($table)) {
        my @fields;
        my $source = '';
        my $target = '';
        if ($type eq 'dropdown') {
            $text = "<SELECT NAME='$field'>\n";
        } elsif ($type eq 'radio-table') {
            @fields = split /\|/, $field_list;
        }

        # Handle junction tables
        if ($table =~ /^(\w+)_to_(\w+)_junction$/) {
            # First see if both tables actually exist
            # Next check to see what we're going from and to
            my ($table1, $table2) = ($1, $2);
            if ($field =~ /^d_$table1\_$self->{_idField}$/i) {
                $target = $table1;
                $source = $table2;
            } elsif ($field =~ /^d_$table2\_$self->{_idField}$/i) {
                $source = $table1;
                $target = $table2;
            }

            my $id = $data->{ucfirst($source) . "\_$self->{_idField}"};
	    my $q = "$table.$source\_$self->{_idField}=$id";
	    if ($query && $query ne '') { $q .= "&&$query"; }
            $sth = $DB->selectRecords("$target<-$table",$q, '');
        } else {
	    my $q = 'rsf=1';
	    
	    if ($query && $query ne '') { $q .= "&&$query"; }
            $sth = $DB->selectRecords($table, $q, 'descriptor');
        }

        while (my $data = $sth->fetchrow_hashref) {
            my $id = $data->{$self->{_idField}} || '';
            my $desc = '';
            if (!$default) {
                $default = $id;
            }
            if ($type eq 'radio-list') {
                $desc = $data->{Description} || '';
            } elsif ($type eq 'radio-table') {
                $text .= " <TR><TD>";
                $desc = "</TD>\n";
                foreach my $field (@fields) {
                    if ($data->{$field} eq '') {
                        $desc .= qq|<TD VALIGN="top">&nbsp;</TD>|;
                    } else {
                        $desc .= qq|<TD VALIGN="top">$data->{$field}</TD>|;
                    }
                }
                $desc .= " </TR>\n";
            } else {
                $desc = $data->{descriptor} || '';
            }
            if ($type =~ /^radio/) {
                $text .= "   <INPUT TYPE='radio' NAME='$field' VALUE='$id'";
                if ($id == $default) { $text .= " CHECKED"; }
                $text .= "> $desc \n";
                if ($type eq 'radio-list') {
                    my $cmd;
                    if (exists $self->{_modules}->{$table.'s'}) {
                        $cmd = $self->cgiCmd($table.'s', 'display', (on=>$id));
                    } else {
                        $cmd = $self->cgiCmd($table, 'display', (on=>$id));
                    }
#                   $text .= qq| <A HREF="$cmd">Details...</A><BR>\n|;
                    $text .= qq| <BR>\n|;
                }
            } elsif ($type eq 'dropdown') {
                $text .= "  <OPTION VALUE='$id'";
                if ($id == $default) { $text .= " SELECTED"; }
                $text .= "> $desc </OPTION>\n";
            }
        }
        if ($type eq 'dropdown') {
            $text .= "</SELECT>\n";
	    if ($table ne 'user' || ! $self->{_CONFIG}->{limit_view_user}) {
            my $cmd = $self->cgiCmd($table.'s', 'display', (on=>''));
            $text .= qq|&nbsp; <A HREF="$cmd">Details...</A><BR>\n|;
	    }
        }
    } elsif ($self->{_DEBUG}) {
        # table wasn't found...  Hmm...
        $text = "Error: table '$table' doesn't exist";
    }
    return $text;
}

=head2 processPage(DB, TEXT, DATA, USER)

Does TT2-style substitutions of HTML widgets and modular commands into $text,
and returns substituted version.

=cut
sub processPage {
    my ($self, $text, $data) = @_;
    $self->trace("processPage") if ($self->{_TRACE});
    my $DB=$self->{_DB};
    my $content = '';

    foreach (split(/\n/, $text)) {

        while (/\[\%\s*dropdown\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/) {
            my ($table, $fieldname, $default) = ($1, $2, $3);
            $fieldname =~ s/^d\_//g;

            if ((defined $data->{$fieldname}) &&
                 $data->{$fieldname} =~ /^\d+/) {
                $default = $data->{$fieldname};
                warn "Setting default for $fieldname to $default\n"
                    if $self->{_DEBUG}>3;
            } else {
                if (! defined $data->{$fieldname}) {
                    warn " data->{$fieldname} is not defined\n" 
                        if $self->{_DEBUG}>3;
                } else {
                    warn " data->{$fieldname} is ".$data->{$fieldname}."\n"
                        if $self->{_DEBUG}>3;
                }
            }
            my $dropdown = $self->makeWidget('dropdown', $data, $table, "d_$fieldname", $default);
            s/\[\%\s*dropdown\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/$dropdown/;
        }

        while (/\[\%\s*dropdown\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\',\s*\'([^\']+)\'\s*\)\s*\%\]/) {
            my ($table, $fieldname, $default, $query) = ($1, $2, $3, $4);
            $fieldname =~ s/^d\_//g;
            if ((defined $data->{$fieldname}) && $data->{$fieldname} =~ /^\d+/) {
                $default = $data->{$fieldname};
            }
            my $dropdown = $self->makeWidget('dropdown', $data, $table, "d_$fieldname", $default, undef, $query);
            s/\[\%\s*dropdown\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'([^\']+)\'\s*\)\s*\%\]/$dropdown/;
        }

        while (/\[\%\s*radio\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/) {
            my $radio = $self->makeWidget('radio', $data, $1, $2, $3);
            s/\[\%\s*radio\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/$radio/;
        }

        while (/\[\%\s*radio-list\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/) {
            my $radio = $self->makeWidget('radio-list', $data, $1, $2, $3);
            s/\[\%\s*radio-list\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/$radio/;
        }

        while (/\[\%\s*radio-table\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w*)\'\,\s*\'([\w\|]+)\'\s*\)\s*\%\]/) {
            my $radio = $self->makeWidget('radio-table', $data, $1, $2, $3, $4);
            s/\[\%\s*radio-table\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w*)\'\,\s*\'([\w\|]+)\'\s*\)\s*\%\]/$radio/;
        }

        # Extra-module commands
        while (/\[\%\s*command\s+(\w+)\:\:(\w+)\s*\%\]/) {
            my $cmd = $self->cgiCmd('stp', $2);
            my $action = qq|<A HREF="$cmd">$2</A>|;
            s/\[\%\s*command\s+(\w+)\:\:(\w+)\s*\%\]/$action/;
        }

        # Intra-module commands
        while (/\[\%\s*command\s+(\w+)\s*\%\]/) {
            my $cmd = $self->cgiCmd($self->{_ARGS}->{modulename}, $1);
            my $action = qq|<A HREF="$cmd">$1</A>|;
            s/\[\%\s*command\s+(\w+)\s*\%\]/$action/;
        }

        if (/\[\%\s*checkbox-list\s*(\w+)\s*(\w+)\s*\%\]/) {
            die trace_tree()."ERROR:  Obsolete checkbox-list syntax found\n";
        }

        # Checkbox lists (requires database lookups)
        while (/\[\%\s*checkbox-list\s*(\w+)\s*(\w+)\s*(\w+)\s*\%\]/) {
            my ($junction, $target, $source) = ($1, $2, $3);  # Hmm, theoretically, $3 can be derived

               warn "checkbox-list loop\n";

            my $id = $data->{$self->{_idField}};
            my $text = '';
            #my $dbh = getDbHandle('eidetic');
            my $sth;
            if ($id =~ /^\d+$/) {
                $sth = $DB->selectRecords("$target<-+$junction", 
                    "$junction.$source\_$self->{_idField}=$id", '');
            } else {
                $sth = $DB->selectRecords("$target", '', '');
            }
            if ($sth) {
                # Create checkboxes
                while (my $item = $sth->fetchrow_hashref) {
                    my $i = $item->{$self->{_idField}} || '';
                    $text .= qq|<INPUT TYPE="checkbox" VALUE="$i" |;
                    my $field_name="$source\_$self->{_idField}";
                    if (defined($item->{$field_name}) &&
                        $item->{$field_name} ne '') {
                            $text .= "CHECKED ";
                    }
                    $text .= sprintf(qq|NAME="d_%s::%s_%s[%d]"> %s<BR>\n|,
                        $junction,
                        $target,
                        $self->{'_idField'},
                        $i,
                        $item->{'descriptor'});
                }
            }
            s/\[\%\s*checkbox-list\s*(\w+)\s*(\w+)\s*(\w+)\s*\%\]/$text/;
        }
        $content .= "$_\n";
    }

    return $content;
}

=head2 do_single

Displays or displays an edit form (HTML) for a single record. Special handling
for admin module. Some TT2-ish stuff.  This is now only used for the display()
and do_userpage() functions.

=cut
sub do_single {
    my ($self, %args) = @_;
    $self->trace("do_single") if ($self->{_TRACE});

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $DB          = $self->{_DB};
    my $on          = $args{'on'};
    my $user        = $args{'user'};
    my $returncontent = $args{'returncontent'};

    my $blank = ($command eq 'edit' ? '' : '&nbsp;');
    
    my $content = '';
    if (-e catfile($self->{_CONFIG}->{applicationTemplates},$modulename,"$command.single.tmpl")) {
	$content = loadFile(catfile($self->{_CONFIG}->{applicationTemplates},$modulename,"$command.single.tmpl"));
    }

    die trace_tree()."Couldn't find $on\n" unless $DB->isInDatabase($table, "$self->{_idField}=$on");

    my $sth = $DB->selectRecords($table, "$self->{_idField}=$on", '') 
	|| die trace_tree()."Couldn't process $on!\n";
    my $data = $sth->fetchrow_hashref;

    my $depth = 1;
    if ($content) {
        while ($content=~/\[\%\s*tree (\w+) (\w+) (\w+) (\d+) (\d+)\s*\%\]/) {
	    my $modulename = $1 || 'folders';
	    my $table = $2 || 'folder';
	    my $parent_field = $3 || 'Parent_uid';
	    my $parent_id = $4 || 0;
	    $depth = $5 || $depth;
            
	    # Determine parent_id by looking up from the database for this
	    # modulename as the resource, and this uid as the resource id,
	    # and then make trees of all of them.
            
	    # We want to select from the database the items where Parent_resource
	    # is projects and Parent_uid = this item's uid.  I may also need to
	    # change the code to determine the top folder by not requiring
	    # uid == Parent_uid.  I also need to consider what happens if
	    # there are multiple top level folders for a project.  If I have a
	    # single folder for the project, should it be hidden?
            
            my $treetxt = '';
            if ($data->{Folder_uid}) {
	        $treetxt = $self->make_tree($modulename, 'display', $table, 
                                            $parent_field, $data->{Folder_uid},
                                            'project', $depth, $user);
            } elsif ($parent_id) {
	        $treetxt = $self->make_tree($modulename, 'display', $table, 
                                            $parent_field, $parent_id,
                                            'project', $depth, $user);
            }

            $content =~ s/\[\%\s*tree (\w+) (\w+) (\w+) (\d+) (\d+)\s*\%\]/$treetxt/;
        }

        my $vars = { %{$user}, %{$data}, depth => $depth };
        $content = $self->substituteValues($content, $vars, $blank);
        $content = $self->parseTemplate($content, $modulename, $table, $data);

        while ($content=~/\[\%\s*subquery\s*(\w+)\:\:(\w+)\s*(\w+)\s*(\w+)\s*([\w=]+)\s*([\!\w]+)\s*(\w+)\s*\%\]/) {
            my $htmlTable = $self->makeHtmlTable($1, $2, $3, $4, 
                                          "$5=".$data->{$self->{_idField}}, $6, $7,
                                          $user);
            $content =~ s/\[\%\s*subquery\s*(\w+)\:\:(\w+)\s*(\w+)\s*(\w+)\s*([\w=]+)\s*([\!\w]+)\s*(\w+)\s*\%\]/$htmlTable/;
        }

    } elsif ($command eq 'display') {
        my @fields = @{ $DB->getTableInfo($table, 'INPUT_FIELDS') };
        $content .= "<TABLE BORDER=0 CELLSPAING=0 CELLPADDING=2>\n";
        $content .= " <TR>\n  <TD WIDTH='150'><B><U>Field</U></B></TD>\n";
        $content .= "  <TD WIDTH='200'><B><U>Value</U></B></TD>\n </TR>\n";
        $content .= " <TR>\n  <TD><B>ID </B></TD>\n  <TD>$data->{uid}</TD>\n </TR>\n";
        foreach my $key (sort @fields) {
            my $value = '&nbsp;';
            if (exists $data->{$key}) {
                $value = $data->{$key};
            }
            $key = ucfirst($key);
            $key =~ s/\_(\w)/ \U$1/g;
            $content .= " <TR>\n  <TD><B>$key</B></TD>\n  <TD>$value</TD>\n </TR>\n";
        }
        $content .= "</TABLE>\n";
    }
    return $content if ($returncontent);
    $self->sendto_screen({title=>ucfirst($data->{descriptor}),
                         content=>$content});

    return ($args{'nextCommand'}, $args{'nextQuery'}, $args{'noticeMsg'});
}

=head2 do_many

Displays or displays edit form or search results for many records. Special 
handling for admin. Some pseudo-TT2 stuff. (HTML)

=cut
sub do_many {
    my ($self, %args) = @_;
    $self->trace("do_many") if ($self->{_TRACE});

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $DB          = $self->{_DB};
    my $on          = $args{'on'};
    my $user        = $args{'user'};

    my $blank = '&nbsp;';
    if ($command eq 'edit') { $blank = ''; }

    die trace_tree(),"Need user object\n" unless defined($user);

    if ($modulename eq 'users' 
	&& defined $user->{Permissions}->{admin}
	&& $user->{Permissions}->{admin} ne 'Y'
        && $self->{_CONFIG}->{limit_view_user}) {
        die trace_tree()."Cannot access user records\n";
    }

    die trace_tree(),"Need table name\n" unless defined($table);

    # A range of items to be edited was requested

    my $rightify = ($on =~ /\!/);
    my $colorify = 1;
    my $newFilter = parseOn($on);

    my $count = 0; ###
    if ($self->{_ARGS}->{query} eq '') { $self->{_ARGS}->{query} = 'all'; }

    if (! $self->{_ARGS}->{sort}) { $self->{_ARGS}->{sort} = $self->{_idField}; }

    die trace_tree(),"Need 'on' parameter\n" unless defined($on);
    die trace_tree(),"Need 'command' parameter\n" unless defined($command);

    my @fields = @{ $DB->getTableInfo($table, 'FIELDS') };
    my $sth = $DB->getfrom_db($table, $on, $self->{_ARGS}->{filter}, $self->{_ARGS}->{sort}, $self->{_ARGS}->{start}, $self->{_ARGS}->{maxitems}*2);
    my $content;
    my @row_text = ();

    my $remainder = $sth->rows() - $self->{_ARGS}->{maxitems};
    if ($remainder < 0) { $remainder = 0; }

    if ($sth == 0) {
        $content = "No available records were found, so we couldn't process the request.";
    } else {
        # TODO: use "catfile" for all of these
        if ($command eq 'edit') {
            $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.list.form.tmpl");
        } elsif ($command eq 'search') {
            $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/display.all.page.tmpl");
        } else {
            $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$self->{_ARGS}->{query}.page.tmpl");
        }
        my $rowTmpl = '';

        if (-e "$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$self->{_ARGS}->{query}.row.tmpl") {
            $rowTmpl = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$self->{_ARGS}->{query}.row.tmpl");
        } elsif (-e "$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.row.tmpl") {
            $rowTmpl = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.row.tmpl");
        } else {
            $rowTmpl = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/display.all.row.tmpl");
        }
        my $color = '';
        my $rowData = '';

        my $count = 0;

        if ($rightify && $sth->rows > $self->{_ARGS}->{maxitems}) {
            $count = $self->{_ARGS}->{maxitems} - $sth->rows;
        }
        while (my $data = $DB->getnext_db($sth)) {
            $count++;
            
            # If hiliting is on, determine what color scheme to use
            if ($self->{_ARGS}->{hilite} &&
                defined($data->{$self->{_ARGS}->{hilite}})) {
# Disabling hiliting feature (not used)
#                if (defined $colorScheme{$data->{$hilite}}) {
#                    $color = $colorScheme{$data->{$hilite}};
#                } else {
#                    # Define new color scheme
#                    $color = nextColor();
#                    $colorScheme{$data->{$hilite}} = $color;
#                }
            } elsif ($color eq "#FFFFFF" && $colorify) {
                $color = "#CCCCCC";
            } else {
                $color = "#FFFFFF";
            }

            # TODO: hack.  see note below
            $data->{row_color} = $color;
            $data->{modulename} = $modulename;
            
            last if ($count > $self->{_ARGS}->{maxitems});
            # Load the template for describing a row in a table, & fill it in
            if ($count > 0) {
                my $rowTmpl2;
                foreach (split /\n/, $rowTmpl) {
                    while (/\[\%\s*command\s+(\w+)\s*\%\]/) {
                        my $cmd = $self->cgiCmd($modulename, $1, (on=>$data->{$self->{_idField}}));
                        my $c = $1;
                        my $action = qq|<A HREF="$cmd">|;
                        if (-e "$self->{_CONFIG}->{globalHtmlRoot}/images/$c.png") {
                            $action .= qq|<img src="/images/$c.png" BORDER="0" ALT="$c"></A>|;
                        } else {
                            $action .= qq|$c</A>|;
                        }
                        s/\[\% command\s+(\w+)\s*\%\]/$action/;
                    }
                    while (/\[\%\s*(\w+)\.(\w+)\s*\%\]/) {
                        my $text = $self->getFieldText($data, $1, $2);

                        if (! $text || $text eq '') {
                            $text = $blank;
                        }
                        s/\[\%\s*$1\.$2\s*\%\]/$text/g;
                    }
                    $rowTmpl2 .= $_ . "\n";
                }
                
                # If there are any sub-tables, load those up too
                if ($rowTmpl2 =~ /\[%\s*\w+\.[\w\.]+/) {
                    $rowTmpl2 =~ s/Host_type_uid\.//g;
                }

                push @row_text, $self->substituteValues($rowTmpl2, $data, $blank) . "\n";
            }
        }

	$rowData = join "\n", @row_text;

        $content =~ s/\[\%\s*row\s*\%\]/$rowData/;
        
        my $sep;
        my $pageCtrl = qq|<P ALIGN="right">|;
        if ( $self->{_ARGS}->{start} > 0) {
            $pageCtrl .= '<A HREF="';
            if ($self->{_ARGS}->{start} <= $self->{_ARGS}->{maxitems}) {
                $pageCtrl .= $self->cgiCmd($modulename, $command, 
                                    (start=>0) );
                $pageCtrl .= "\">First Page</A> ";
            } else {
                $pageCtrl .= $self->cgiCmd($modulename, $command, 
                                    (start=>$self->{_ARGS}->{start}-$self->{_ARGS}->{maxitems}) );
                $pageCtrl .= "\">Prev $self->{_ARGS}->{maxitems}</A> ";
            }
            $sep = " | ";
        } else {
            $sep = " ";
        }
        if ($remainder > 0 && ($count >= $self->{_ARGS}->{maxitems} || $rightify)) {
            $pageCtrl .= $sep;
            $pageCtrl .= '<A HREF="';
            $pageCtrl .= $self->cgiCmd($modulename, $command, 
                                (start=>$self->{_ARGS}->{start}+$self->{_ARGS}->{maxitems}) );
            
            if ($remainder < $self->{_ARGS}->{maxitems}) {
                $pageCtrl .= "\">Next $remainder...</A>";
            } else {
                $pageCtrl .= "\">Next $self->{_ARGS}->{maxitems}...</A>";
            }
        } elsif ($self->{_ARGS}->{start} < 1) {
	    $pageCtrl = '';
	}
	if ($pageCtrl ne '') {
	    $pageCtrl .= "</P>\n";
	}
        $content = $pageCtrl . $content . $pageCtrl;
        
        my %var = %$user;
        $var{'pageCtrl'} = $pageCtrl;

        if (-e "$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$self->{_ARGS}->{query}.header.tmpl") {
            $var{header} = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$self->{_ARGS}->{query}.header.tmpl");
        } elsif (-e "$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.header.tmpl") {
            $var{header} = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.header.tmpl");
        } else {
            $var{header} = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/display.all.header.tmpl");
        }

        # TODO: these feel like hacks.  Maybe move to sendto_screen?
        $var{'on'} = $on;
        $var{'modulename'} = $modulename;
	
        $content = $self->substituteValues($content, \%var, $blank);
    }
    my $title = ucfirst($self->{_ARGS}->{query});
    $title =~ s/\_(\w)/ \U$1/g;

    $self->sendto_screen({title=>$title, content=>$content});

    return ($args{'nextCommand'}, $args{'nextQuery'}, $args{'noticeMsg'});
}

=head2 make_tree

Returns a tree content object

modulename - the name of the tree-able data, such as folders.  This module must
  support parent/child relationships within itself.

command - used to indicate what type of templates will be used.

table - the table associated with modulename

DB - the database object created via new Eidetic::Database

parent_field - the name of the data field used to map the tree-able data into a
  tree.  

parent_id - the id of the top parent object

parent_resource - the name of the resource (table) owning the root item(s)

parent_resource_id - the id of the resource (record uid) owning the root item(s)

depth - the number of levels of the tree to decend down to.

user - the user object, or undef to have generic settings.

=cut
sub make_tree {
    # TODO:  Change to use hash for args?
    my $self = shift;
    $self->trace("make_tree") if ($self->{_TRACE});
    my $DB=$self->{_DB};
    my $modulename = shift      || die trace_tree() . "No modulename specified\n";
    my $command = shift         || die trace_tree() . "No command specified\n";
    my $table = shift           || die trace_tree() . "No table specified\n";
    my $parent_field = shift    || die trace_tree() . "No parent field specified\n";
    my $parent_id = shift       || die trace_tree() . "No parent id specified\n";
    my $parent_resource = shift || die trace_tree() . "No parent resource specified\n";
    my $depth = shift || 3;
    my $user = shift;  
    my $no_resources = shift || 0;
    my $on = shift;
    my $old_parent_uid = shift;

    if ($self->{_TRACE}) {
        warn "make_tree(\n";
        warn "  modulename      = $modulename\n";
        warn "  command         = $command\n";
        warn "  table           = $table\n";
        warn "  parent_field    = $parent_field\n";
        warn "  parent_id       = $parent_id\n";
        warn "  parent_resource = $parent_resource\n";
        warn "  depth           = $depth\n";
        warn "  user            = $user\n";
        warn "  no_resources    = $no_resources\n";
        warn ")\n";
    }

    die "Depth argument '$depth' to make_tree() is not valid number\n" 
        unless ($depth =~ /^\d+$/);

    my $blank = '&nbsp;';
    if ($command eq 'edit') { $blank = ''; }

    if (! $self->{_ARGS}->{sort}) {   # Sort by parent?
        $self->{_ARGS}->{sort} = $self->{_idField}; 
    }

    my @fields = @{ $DB->getTableInfo($table, 'FIELDS') };
    my $content;
    my @row_text = ();
    my @rows = ();

    # What we need to do is trace down all components in the given folder
    # $parent_id and get them all in a big ol' list.  This query could span
    # a range of different tables in the database.  We decend to $depth
    # level.

    my $q = $self->{_ARGS}->{query} || 'all';

    # Load template file for this operation
    $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$q.tree.page.tmpl");

    # Load header
    my $header = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$q.tree.header.tmpl");

    # Load footer
    my $footer = '';
    my $footer_file = "$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$q.tree.footer.tmpl";
    if (-e $footer_file) {
	$footer = loadFile($footer_file);
    }

    # Substitute header & footer
    $header =~ s/\[\%\s*$self->{_idField}\s*\%\]/$parent_id/g;
    $footer =~ s/\[\%\s*$self->{_idField}\s*\%\]/$parent_id/g;
    $content =~ s/\[\%\s*footer\s*\%\]/$footer/;
    $content =~ s/\[\%\s*header\s*\%\]/$header/;

    # Load folder & row templates
    my $folderTmpl = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$q.tree.folder.tmpl");
    my $rowTmpl = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$q.tree.row.tmpl");

    # Look up parent ids for this record
    my $pth = $DB->selectRecords($table, "Parent_resource='$parent_resource'&Parent_uid=$parent_id", $self->{_ARGS}->{sort});

    my @parents = ($parent_id);
    my $parent_of_root;

    my $img_url = $self->{_CONFIG}->{images_url} || '';
    my $icon_folderopen   = qq(<img src="$img_url/tigris/icon_arrowfolderopen2_sml.gif" border="0" alt="[_]">);
    my $icon_folderclosed = qq(<img src="$img_url/tigris/icon_arrowfolderclosed1_sml.gif" border="0" alt="[+]">);
    my $icon_document = qq(<img src="$img_url/tigris/icon_doc_sml.gif" border="0" alt="O">);
    my $icon_spacer = qq(<img src="$img_url/blank.gif" width="16" height="1" border="0" alt="">);

    # TODO:  Reorganize the following block to separate the processing from
    # the data retrieval, so it can make multiple db calls based on the depth
    for (my $level = 0; ($level<=$depth) && ($level<$self->{_CONFIG}->{max_folder_depth}); $level++) {
	my $pfilter = '';
	last if ($#parents < 0);  # No parents left
	if ($level == 0) {
	    # This is the root node
	    $pfilter = "$self->{_idField}=$parent_id";
	} elsif ($#parents == 0) {
	    # There is only one parent
	    $pfilter = "$parent_field=$parents[0]&&$self->{_idField}<>$parent_id";
	} else {
	    # There are multiple parents to select
	    $pfilter = "$parent_field IN (" . join(',', @parents) . ")"
		. "&&$self->{_idField}<>$parent_id";
	}

	warn "Sorting by $self->{_ARGS}->{sort}\n" if ($self->{_DEBUG}>2);

	# Select all child items that have parent == $parent_id
	my $sth = $DB->getfrom_db($table, '', $pfilter, $self->{_ARGS}->{sort}, 0);
	@parents = ();

	last if ($sth == 0);

	# Foreach record retrieved
        while (my $data = $DB->getnext_db($sth)) {
	    # Note the current level in the {symbol} field
	    $data->{level} = $level;
	    if ($level == $depth) {
		$data->{symbol} = ($icon_spacer x $level) . $icon_folderclosed;
	    } else {
		$data->{symbol} = ($icon_spacer x $level) . $icon_folderopen;
	    }
            $data->{row_color}      = '#FFFFFF';  # TODO:  Consider colorizing?
            $data->{modulename}     = $modulename;
            $data->{images_url}     = $self->{_CONFIG}->{images_url};
            $data->{Resource_name}  = $data->{Parent_resource};
            $data->{Resource_uid}   = $data->{$self->{_idField}};
            $data->{Old_parent_uid} = $old_parent_uid;
            $data->{on}             = $on;
            $data->{Top}            = $parent_id;

            # Load the template for describing a row in a table, & fill it in
	    my $folderTmpl2;

	    # It would be desireable to move the following block to template(s)
	    foreach (split /\n/, $folderTmpl) {
		# Substitute cgiCmd
		while (/\[\%\s*command\s+(\w+)\s*\%\]/) {
		    my $cmd = $self->cgiCmd($modulename, $1, (on=>$data->{$self->{_idField}}));
		    my $c = $1;
		    my $action = qq|<a href="$cmd">|;
		    if (-e "$self->{_CONFIG}->{globalHtmlRoot}/images/$c.png") {
			$action .= qq|<img src="$img_url/images/$c.png" border="0" img src="$c"></A>|;
		    } else {
			$action .= qq|$c</A>|;
		    }
		    s/\[\% command\s+(\w+)\s*\%\]/$action/;
		}
		# Substitute x.y data
		while (/\[\%\s*(\w+)\.(\w+)\s*\%\]/) {
		    my $text = $self->getFieldText($data, $1, $2);
		    
		    if (! $text || $text eq '') {
			$text = $blank;
		    }
		    s/\[\%\s*$1\.$2\s*\%\]/$text/g;
		}
		$folderTmpl2 .= $_ . "\n";
	    }

	    my $pre_text = '';
	    if ($level == 0 
                && $command ne 'move'
		&& $data->{Parent_uid}
		&& $data->{Parent_uid} != $data->{$self->{_idField}}) {

		# Add the parent folder (i.e., '..')
		$pre_text = $self->substituteValues($folderTmpl, {
		    row_color  => '#FFFFFF',
		    descriptor => '<b>..</b>',
		    uid        => $data->{Parent_uid},
                    modulename => $data->{Parent_resource} . 's',
		    depth      => $depth,
		}, $blank, 'tt2');
	    }

	    $data->{depth} = $depth;
	    $data->{row_data} = $pre_text 
		. $self->substituteValues($folderTmpl2, $data, $blank, 'tt2') . "\n";

	    # Add record to list
	    push @rows, $data;

	    if ($level == 0) {
		@parents = ($data->{uid});
	    } elsif ($data->{$self->{_idField}} != $data->{$parent_field}) {
		# Add uid to the next @parents list
		push @parents, $data->{$self->{_idField}};
	    }
	}
    }

    # Handle if we wish to reorganize the data into a tree form
    my @tree = list_to_tree(\@rows, $self->{_idField}, $parent_field);

    my $row_data = '';
    # TODO:  What if we want to sort folders by something other than their name?
    foreach my $row (tree_to_list(\@tree,['descriptor'],[\&alphabetically])) {
	next unless (defined $row->{$self->{_idField}});

	$row_data .= $row->{row_data};
	$row_data .= "\n";

	# Next go through each folder we've loaded and get its resources
	# select records where Folder_uid IN @folder_list sort by Folder_id

	next if (!defined $row->{level} || $row->{level} == $depth || $no_resources);

	my $dfilter = "folder_to_resource.Folder_uid=".$row->{$self->{_idField}};
	warn "Sorting docs by $self->{_ARGS}->{sort}\n" if ($self->{_DEBUG}>1);
	my $sth = $DB->selectRecords('document<-folder_to_resource', $dfilter, $self->{_ARGS}->{sort}, 0);
	while (my $item = $sth->fetchrow_hashref) {
	    $item->{Parent_resource} = 'folder';
            $item->{Top} = $parent_id;
	    $item->{symbol} = ($icon_spacer x (1+$row->{level})) . $icon_document;
            $item->{row_color}  = '#FFFFFF'; 
            $item->{modulename} = $modulename;
            $item->{images_url} = $self->{_CONFIG}->{images_url};
	    if ($item->{Alias}) {
		$item->{descriptor} = $item->{Alias};
	    }

	    # Now we know the item's resource type (i.e., table) and id
	    # We could query for data about this item at this point... 

	    # It would be desireable to move the following block to template(s)
	    my $rowTmpl2;
	    foreach (split /\n/, $rowTmpl) {
		# Substitute cgiCmd
		while (/\[\%\s*command\s+(\w+)\s*\%\]/) {
		    my $cmd = $self->cgiCmd($modulename, $1, (on=>$item->{$self->{_idField}}));
		    my $c = $1;
		    my $action = qq|<A HREF="$cmd">|
                        . qq|<img src="$img_url/$c.png" border="0" alt="$c"></A>|;
		    s/\[\% command\s+(\w+)\s*\%\]/$action/;
		}
		# Substitute x.y data
		while (/\[\%\s*(\w+)\.(\w+)\s*\%\]/) {
		    my $text = $self->getFieldText($item, $1, $2);
		    
		    if (! $text || $text eq '') {
			$text = $blank;
		    }
		    s/\[\%\s*$1\.$2\s*\%\]/$text/g;
		}
		$rowTmpl2 .= $_ . "\n";
	    }

	    $row_data .= $self->substituteValues($rowTmpl2, $item, $blank, 'tt2') . "\n";
	}

    }
    $content =~ s/\[\%\s*row\s*\%\]/$row_data/;
        
    my %var = (
	       parent_id        => $parent_id,
	       Parent_uid       => $parent_id,
	       Parent_resource  => 'folders',
               images_url       => $self->{_CONFIG}->{images_url},
	       depth            => $depth
	       );

    return $self->substituteValues($content, \%var, $blank, 'tt2');
}

=head2 do_tree

Displays a collection of data in tree form.

=cut
sub do_tree {
    my ($self, %args) = @_;
    $self->trace("do_tree") if ($self->{_TRACE});

    my $modulename   = $args{'modulename'};
    my $command      = $args{'command'};
    my $table        = $args{'table'};
    my $DB           = $self->{_DB};
    my $parent_field = $args{'parent_field'};
    my $parent_id    = $args{'parent_id'};
    my $depth        = $args{'depth'};
    my $user         = $args{'user'};

    die "Depth argument '$depth' to do_tree() is not valid number\n"
        unless ($depth =~ /^\d+$/);

    my $title = ucfirst($self->{_ARGS}->{query});
    $title =~ s/\_(\w)/ \U$1/g;

    my $parent_resource = 'folder';

    my $content = $self->make_tree($modulename, $command, $table, 
                                   $parent_field, $parent_id, $parent_resource, 
                                   $depth, $user);
    $self->sendto_screen({title=>$title,
                          content=>$content,
                          depth => $depth,
                          Notice => $args{'noticeMsg'}});

    return ($args{'nextCommand'}, $args{'nextQuery'}, $args{'noticeMsg'});
}


=head2 get_owner($resource, $resource_id, $owner_field)

Returns the name of the owner of the given resource
Dies if database is not usable or if $owner_field not defined for $resource
Returns 0 if resource_id cannot be found in table $resource 
Returns 'unknown' if it cannot find matching username for user_id listed
with the given resource item.

$resource - the name of the table to look in ('folder', 'document', etc.)
$resource - the id of the item to test ownership of
$owner_field - the name of the field containing the owner id number

=cut
sub get_owner {
    my ($self, $resource, $resource_id, $owner_field) = @_;
    $self->trace("get_owner") if ($self->{_TRACE});
    # Verify that the user owns the item to be moved

    warn ("Retrieving $self->{_idField}=$resource_id from $resource") if ($self->{_DEBUG}>2);
    my $sth = $self->{_DB}->selectRecords($resource, "$self->{_idField}=$resource_id", '');
    die "Could not retrieve from database" if ($sth == 0);
    return 0 if ($sth->rows == 0);

    my $item = $sth->fetchrow_hashref;

    die "'$owner_field' field not defined for $resource table" 
	unless ( defined $item->{$owner_field} );

    my $owner = $self->{_DB}->selectRecords('user', 
				       'uid='.$item->{$owner_field}, 
				       '');
    if ($owner != 0) {
            return ($owner->fetchrow_hashref)->{descriptor};
    } else {
            return 'unknown';
    }
}

=head2 has_access()

This subroutine determines if a given user has access to perform a given
operation on a particular record in a table.  For instance, it determines
if the user is the owner of a document, or if the document is project-owned,
if they are a member of that project.

   $user - the hashref for the person to check access on
   $item - the hashref for the data row
   $owner_field - the name of the column in the $item hashref that indicates 
           the id of the user that owns it
   $project_field - the name of the column in the $item hashref that indicates
           the id of the project that owns it

=cut

sub has_access {
    my $self = shift;
    my $DB=$self->{_DB};
    my $user = shift || die "Invalid user object passed to has_access()\n";
    my $item = shift;
    my $owner_field = shift || '';
    my $project_field = shift || '';

    warn "has_access():  Checking access permission for user " . $user->{uid} . "\n"
        if ($self->{_DEBUG}>2);

    if (! defined $user->{Permissions}) {
        warn "User has no permissions defined\n" 
            if ($self->{_DEBUG}>2);
    }

    # Admin's have full access to everything
    if (defined $user->{Permissions}->{admin} 
        && $user->{Permissions}->{admin} eq 'Y') {
        warn "User is an admin\n" 
            if ($self->{_DEBUG}>2);
        return 'admin';
    }

    # If no object was specified, default to no access
    if (! $item) {
        warn "No object specified to has_access()\n" 
            if ($self->{_DEBUG}>2);
        return 0;
    }

    # Is user owner of the item?
    return 'owner' if ($owner_field && $item->{$owner_field} == $user->{$self->{_idField}});

    # Project ownership active?
    if (!($project_field && exists $item->{$project_field})) {
        warn "Project ownership not active in has_access()\n" 
            if ($self->{_DEBUG}>2);
        return 0;
    }
    if (! ($DB->checkTableExistance('project_to_user_junction'))) {
        warn "project_to_user_junction table does not exist\n"
            if ($self->{_DEBUG}>2);
        return 0;
    }

    # Is item project owned, and is user member of that project?
    my $sth = $DB->selectRecords('project_to_user_junction', 
				 "project_uid=" . $item->{$project_field}
				 ."&&user_uid=" . $user->{$self->{_idField}}
				 ."&&rsf>0", 
				 '');

    die "Could not retrieve data from project_to_user_junction table" if ($sth == 0);
    if ($sth->rows == 0) {
        warn "No records retrieved in has_access()\n"
            if ($self->{_DEBUG}>2);
        return 0;
    }
    my $project = $sth->getrow_hashref;

    if (! ($project && defined $project->{role})) {
        warn "project role not defined in has_access()\n"
            if ($self->{_DEBUG}>2);
        return 0;
    }

    # If we got this far, the user has project access
    return 'project ' . $project->{role};
}

=head2 do_move

Moves an item within a tree structure

    $modulename - type of thing to be moved
    $on         - identity of thing to be moved
    $dataHandle - data structure expecting the fields:
                  Parent_resource
                  Old_parent_uid
                  New_parent_uid

The user must be the owner of the document to be moved.

=cut
sub do_move {
    my ($self, %args) = @_;
    $self->trace("do_move") if ($self->{_TRACE});

    my $modulename     = $args{'modulename'};
    my $command        = $args{'command'};
    my $table          = $args{'table'};
    my $DB             = $self->{_DB};
    my $on             = $args{'on'};
    my $data           = $args{'data'};
    my $emails         = $args{'emails'};
    my $nextCommand    = $args{'nextCommand'};
    my $nextQuery      = $args{'nextQuery'};
    my $user           = $args{'user'};
    my $noticeMsg      = $args{'noticeMsg'};
    my $nextOn         = $on;
    my $nextModulename = $modulename;

    die trace_tree()."Error - table name not given for $command $modulename (on $on)\n" 
        unless $table;

    die "Unknown owner id\n" unless (defined $user->{$self->{_idField}});
    my $owner = $self->get_owner($table, $on, 'Submitter_uid_user');

    my $content = '';
    my $var = { };
    if (!$DB->isInDatabase($table, "$self->{_idField}=$on")) {
        die "Couldn't find record number $on in table $table\n";
    } elsif ($owner) {
        $content = "This is owned by $owner; only they can move it.\n";
        $self->sendto_screen({title=>"Move",
                              content=>$content,
                              %$var});
    } elsif (! $data->{Old_parent_uid}) {
        die "Error:  Must indicate where to move from\n";
    } elsif (! $data->{New_parent_uid}) {
        $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/move.tmpl");
        my $sth = $DB->selectRecords($table, 
				     $self->{_idField}.'='.$on, 
				     '')
	    || die trace_tree()."Couldn't process $on!\n";
        $var = { %{$sth->fetchrow_hashref}, %{$data} };

        # We want to display as much of the tree as we can, so set depth to 8
        my $depth = 8;
        $data->{Top} ||= $data->{Old_parent_uid};

        $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/move.tmpl");
        $content .= $self->make_tree($modulename, 'move', 'folder',
                                     'Parent_uid', $data->{Top},
                                     'folder', $depth, $user, 1, $on,
                                     $data->{Old_parent_uid});
        $self->sendto_screen({title=>"Move",
                              content=>$content,
                              %$var});
 
    } elsif ($modulename eq 'documents') {
        # Relink the document
        $self->{_subobj}->create_document_link(
                                               $on,
                                               $data->{New_parent_uid}
                                               ) || carp "Could not attach file: $!\n";
        $self->{_subobj}->remove_document_link(
                                               $on,
                                               $data->{Old_parent_uid}
                                               ) || carp "Could not remove link: $!\n";
        $noticeMsg = "Document has been successfully moved.\n";
        $nextCommand = 'display';
        $nextModulename = 'folders';
        $nextOn = $data->{New_parent_uid}.'[3]';
    } elsif ($modulename eq 'folders') {
        if ($data->{New_parent_uid} == $on) {
            die "Sorry, you cannot move a folder into itself.";
        } else {
            # Move the folder
            $self->trace("Moving folder $on\n") if ($self->{_TRACE});
            $self->{_subobj}->move_folder($on, 
                                          $data->{New_parent_uid}, 
                                          $user->{$self->{_idField}})
                || carp "Could not move folder: $!\n";
            $noticeMsg = "Folder has been successfully moved.\n";
            $nextCommand = 'display';
            $nextModulename = 'folders';
            $nextOn = $data->{New_parent_uid}.'[3]';
        }
    }

    return ($nextCommand, $nextQuery, $noticeMsg, $nextOn, $nextModulename);
}


=head2 do_attach

Attaches a document to a folder

    $modulename - type of thing to be moved
    $on         - identity of thing to be moved
    $dataHandle - data structure expecting the fields:
                  Parent_resource
                  New_parent_uid

The user must be the owner of the document to be attached.

Note that this is valid only for documents, not folders

=cut
sub do_attach {
    my ($self, %args) = @_;
    $self->trace("do_attach") if ($self->{_TRACE});

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $DB          = $self->{_DB};
    my $on          = $args{'on'};
    my $dataHandle  = $args{'dataHandle'};
    my $emails      = $args{'emails'};
    my $nextCommand = $args{'nextCommand'};
    my $nextQuery   = $args{'nextQuery'};
    my $user        = $args{'user'};
    my $noticeMsg   = $args{'noticeMsg'};

    die trace_tree()."Error - table name not given for $command $modulename (on $on)\n" 
        unless $table;

    my $owner = $self->get_owner($table, $on, 'Submitter_uid_user');

    my $var = { };
    my $content = '';
    if ($owner ne $user->{descriptor} && $user->{Permissions}->{admin} ne 'Y') {
	$content = "This is owned by $owner; only they can attach it here.\n";
    } elsif (! defined $dataHandle->{New_parent_uid}) {
	$content = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/attach.tmpl");
	my $sth = $DB->selectRecords($table, 
				     'uid='.$on, 
				     '');
	$var = { %{$sth->fetchrow_hashref}, %{$dataHandle} };
	
    } else {
	# Attach the item
	warn "Parent Uid: ", $dataHandle->{New_parent_uid}, "\n" if ($self->{_DEBUG}>2);
	$self->{_subobj}->create_document_link($on,
				   $dataHandle->{New_parent_uid}
				   ) || carp "Could not attach file: $!\n";
	$content = "Attachment successfully made";
    }

    $self->sendto_screen({title=>"Attach", content=>$content, %$var});
    return ($nextCommand, $nextQuery, $noticeMsg);
}

=head2 do_detach

=cut

sub do_detach {
    my ($self, %args) = @_;
    $self->trace("do_detach") if ($self->{_TRACE});

    # TODO:  Remove a link to a document (derive from do_move?)
    # TODO:  Make do_remove remove all links to the doc, too

    my $content = 'Detach is not yet implemented';

    $self->sendto_screen({title=>"Detach", content=>$content});
    return ($args{nextCommand}, $args{nextQuery}, $args{noticeMsg});
}

=head2 do_give

Generates form to allow changing of ownership of a document to a new user

=cut

sub do_give {
    my ($self, %args) = @_;
    $self->trace("do_give") if ($self->{_TRACE});

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $DB          = $self->{_DB};
    my $on          = $args{'on'};
    my $data        = $args{'data'};
    my $user        = $args{'user'};
    my $nextCommand = $args{'nextCommand'};
    my $nextQuery   = $args{'nextQuery'};
    my $noticeMsg   = $args{'noticeMsg'};

    die "No document id number specified to be given\n" unless ($on =~ /^\d+$/);

    # Get the current owner of the document
    my $owner = $self->get_owner($table, $on, 'Submitter_uid_user');

    die "Could not determine owner for $table ($on)\n" unless ($owner);

    my $content = '';

    # Current owner must be the same as the current user
    if ($owner eq $user->{descriptor}
	|| $user->{Permissions}->{admin} eq 'Y') {

	die trace_tree()."Couldn't find $on\n" unless $DB->isInDatabase($table, "$self->{_idField}=$on");

	my $sth = $DB->selectRecords($table, "$self->{_idField}=$on", '')
	    || die trace_tree()."Couldn't process $on!\n";
	$data = $sth->fetchrow_hashref;

	$content = loadFile(catfile($self->{_CONFIG}->{applicationTemplates},$modulename,"give.tmpl"));
    } else {
        $content = "Ownership of this document can only be transferred "
            . "by its current owner ($owner).\n";
	warn ("Attempt by $user->{$self->{_idField}} to give document owned by '$owner'") if ($self->{_DEBUG}>1);
    }

     $self->sendto_screen({
         title     => 'Give',
         Notice    => $noticeMsg,
         content   => $content,
         %$data,
     });
    return ($nextCommand, $nextQuery, $noticeMsg);
}


=head2 do_remove_confirm

=cut

sub do_remove_confirm {
    my ($self, %args) = @_;
    $self->trace("do_remove_confirm") if ($self->{_TRACE});

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $DB          = $self->{_DB};
    my $on          = $args{'on'};
    my $data        = $args{'data'};
    my $nextCommand = $args{'nextCommand'};
    my $nextQuery   = $args{'nextQuery'};
    my $noticeMsg   = $args{'noticeMsg'};

    my $content;

    if ($on !~ /^\d+$/) {
        $noticeMsg = "You must specify a valid ID to remove.";
    } else {
        if (!$DB->isInDatabase($table, "$self->{_idField}=$on")) {
            die trace_tree()."Couldn't find $on\n";
        }

        # Load template file for this operation
        $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/$modulename/remove_confirm.tmpl");

        my $sth = $DB->selectRecords($table, $self->{_idField}."=$on", "")
            || die trace_tree()."Couldn't process $on!\n";

        $data = $sth->fetchrow_hashref;
	
    }
    $self->sendto_screen({title=>'Please confirm remove',
                         content=>$content, %$data});

    return ($nextCommand, $nextQuery, $noticeMsg);
}

=head2 do_remove

Removes items from table (by setting rsf=0; does not actually delete anything.)

=cut
sub do_remove {
    my ($self, %args) = @_;
    $self->trace("do_remove") if ($self->{_TRACE});

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $DB          = $self->{_DB};
    my $on          = $args{'on'};
    my $data        = $args{'data'};
    my $emails      = $args{'emails'};
    my $nextCommand = $args{'nextCommand'};
    my $nextQuery   = $args{'nextQuery'};
    my $user        = $args{'user'};
    my $noticeMsg   = $args{'noticeMsg'};

    my $content;
    my $user_id = $user->{user_id} || 0;

    die "You must specify a valid ID to remove.\n" unless ($on =~ /^\d+$/); 

    # Catch if user doesn't have access to remove the item
    my $owning_user_field = '';
    my $owning_project_field = '';
    if ($modulename eq 'documents') {
	$owning_user_field = 'Submitter_uid_user';
	$owning_project_field = 'Group_uid_project';
    }

    # Retrieve document record
    my $sth = $DB->selectRecords($table, "$self->{_idField}=$on", '')
	|| die trace_tree()."Couldn't process $on!\n";
    my $item = $sth->fetchrow_hashref;

    die "Couldn't find record in database to remove\n" unless ($item);

    if (!$self->has_access($user, $item, $owning_user_field, $owning_project_field)) {
	$noticeMsg =  "You do not have access permission for removing this item.\n"
            . "Please contact the owner of the item or your administer.\n";
    } elsif (!$DB->isInDatabase($table, "$self->{_idField}=$on")) {
	$noticeMsg = "Error:  Couldn't find item number $on in database to remove.\n";
    } elsif ($modulename eq 'folders' 
	     && ($DB->isInDatabase('folder', "$self->{_idField}<>$on&&Parent_uid=$on&&rsf>0")
		 || $DB->isInDatabase('folder_to_resource', "Folder_uid=$on&&rsf>0"))) {
	$noticeMsg = "Folder isn't empty, so you cannot delete it.\n";
    } else {
        my $id = $on;
        if (!$DB->isInDatabase($table, "$self->{_idField}=$id")) {
            die trace_tree()."Couldn't find $id\n";
        }

        $data->{$self->{_idField}} = $id;
        $data->{rsf} = 0;
        $DB->updateRecord($modulename, $user_id,
            $self->{_modules}->{$modulename}->{tables},
            %$data);

        if ($modulename eq 'documents') {
            my $cmd = "DELETE FROM folder_to_resource WHERE Resource_uid=$id";
            my $sth = $DB->{_handle}->prepare($cmd)
                or confess "Cannot prepare database command:  " . $DB->{_handle}->errstr;
            $sth->execute;
        }

        $table =~ s/\_/ /g;
        $noticeMsg = "$table entry $on was successfully removed.\n";
    }
    $content = $noticeMsg;
    $self->sendto_screen({title=>'Remove successful',
                         content=>$content, %$data});
    return ($nextCommand, $nextQuery, $noticeMsg);
}


sub parseTemplate {
    my ($self, $tmpl, $modulename, $table, $data) = @_;
    $self->trace("parseTemplate") if ($self->{_TRACE});
    my $DB=$self->{_DB};
    my $output = '';

    foreach (split /\n/, $tmpl) {
        while (/\[\%\s*command\s+(\w+)\s*\%\]/) {
            my $cmd = $self->cgiCmd($modulename, $1, (on=>$data->{$self->{_idField}}));
            my $c = $1;
            my $action = qq|<A HREF="$cmd">|;
            if (-e "$self->{_CONFIG}->{globalHtmlRoot}/images/$c.png") {
                $action .= qq|<img src="/images/$c.png" border="0" alt="$c"></A>|;
            } else {
                $action .= qq|$c</A>|;
            }
            s/\[\% command\s+(\w+)\s*\%\]/$action/;
        }
        while (/\[\%\s*(\w+)\.(\w+)\s*\%\]/) {
            my $text = $self->getFieldText($data, $1, $2);
            
            s/\[\%\s*$1\.$2\s*\%\]/$text/g;
        }
        while (/\[\%\s*junction\s*(\w+)\s*(\w+)\s*\%\]/) {
            my ($junction, $target, $source) = ($1, $2, $table);

            my $id = $data->{$self->{_idField}};
            my $sth = $DB->selectRecords("$target<-$junction",
				    "$junction.$source\_$self->{_idField}=$id",
                                    "$junction.$target\_$self->{_idField}");

            my $text;
            while (my $data = $sth->fetchrow_hashref) {
                my $cmd = '';
                if (exists $self->{_modules}->{$target.'s'}) {
                    $cmd = $self->cgiCmd($target.'s', 'display', (on=>$data->{$self->{_idField}}));
                } else {
                    $cmd = $self->cgiCmd($target, 'display', (on=>$data->{$self->{_idField}}));
                }
                $text .= qq|<A HREF="$cmd">|;
                $text .= $data->{descriptor} . "</A><BR>\n";
            }
            s/\[\%\s*junction\s*$junction\s*$target\s*\%\]/$text/;
        }
        $output .= $_ . "\n";
    }

    return $self->substituteValues($output, $data) . "\n";
}


########################################################################
## MAIL SUBSYSTEM                                                     ##
########################################################################

=head1 Mail subsystem

=head2 sendto_email(TO, TITLE, CONTENT)

Sends mail to $to, by a pipe to the mail command. Eidetic::Mail uses the 
Mail::Sendmail module, which works better.

=cut
sub sendto_email {
    my ($self, $to, $title, $content) = @_;
    $self->trace("sendto_email") if ($self->{_TRACE});

    my $sender = new Mail::Sender({ smtp => "localhost" });
    die trace_tree()."could not start email: ".$Mail::Sender::Error."\n"
        unless(defined($sender));

    if (!ref($sender->MailMsg({
        to => $to,
        from => $self->{_CONFIG}->{'webmaster'},
        subject => $title,
        msg => $content,
        }))) {
        die trace_tree()."could not send email: ".$sender->{'error_msg'}."\n";
    }
    return 1;
}


=head2 notifyStaff(MODULE, DATA)

Sends %data to the hardcoded staff addresses via sendto_email.

=cut
# TODO: move template junk into sendto_email
# notifyStaff
sub notifyStaff {
    my ($self, $modulename, %data) = @_;
    $self->trace("notifyStaff") if ($self->{_TRACE});
    my ($address, @addresses, $key, @keys, @lines, $sep);
    my $intro = "$modulename entry has been submitted:\n\n";
    my $subject = "$modulename submission";

    my @addrlist;

    # Handle Lab project emails (TODO: this should come from the DB)
    if (ref($self->{_CONFIG}->{project_notification_emails}) eq "ARRAY") {
        $addrlist[1] = $self->{_CONFIG}->{project_notification_emails};
    }
    else {
        $addrlist[1] = [ $self->{_CONFIG}->{project_notification_emails} ];
    }

    if (ref($self->{_CONFIG}->{project_notification_emails_jp}) eq "ARRAY") {
        $addrlist[2] = $self->{_CONFIG}->{project_notification_emails_jp};
    }
    else {
        $addrlist[2] = [ $self->{_CONFIG}->{project_notification_emails_jp} ];
    }

    @addresses = $data{Lab_uid} eq '1' ? @{$addrlist[1]} : @{$addrlist[2]};
    if (defined $data{Coordinator_email} && $data{Coordinator_email} ne '') {
        push @addresses, $data{Coordinator_email};
    }
    
    @keys = sort keys %data;

    my $message = '';

    my $templateFile = catfile($self->{_CONFIG}->{'applicationTemplates'},
                               $modulename,"notifyStaff.tmpl");
    if (-f $templateFile) {
        # Load template
        my $template = loadFile($templateFile);
        $self->{_template}->process(\$template, \%data, \$message)
            || die "Could not process template $templateFile: ",
                $self->{_template}->error(), "\n";

    } else {
        $message .= $intro;
        foreach $key (@keys) {
            my $value = '';
            if ($data{$key}) {
                $value = join "\n", wrap('  ', '  ', $data{$key});
            }
            $sep = '-' x (length $key) ;
            $message .= $key . "\n" . $sep . "\n" . $value . "\n\n";
        }
    }

    my $to = join ', ', @addresses;
    $self->sendto_email($to, $subject, $message);

    return "";
}

########################################################################
## UTILITIES                                                          ##
########################################################################

=head1 Utility functions

=head2 handleError(MODULE, ERROR)

Prints error messages.

=cut
# handleError
sub handleError {
    my ($self, $modulename, $error) = @_;
    $self->trace("handleError") if ($self->{_TRACE});
    my ($result, @comments) = split(/ /, $error);
    my ($errmsg);

    if (!defined $result || $result =~ /^success/i || $result =~ /^message/i || $result =~ /^note/i || $result eq '') {
        # do nothing; there was no error
        return '';
    } elsif ($result eq 'failure' || $result eq 'error') {
        $errmsg = join(' ', @comments);
    } else {
        # This may have been an internal exception
        $result = "Error";
        $errmsg = $error;
    }
    #$errmsg =~ s/at [\w\.]+ line \d+//sg;

    # I need a way to insert $reason and @comments into the page.

    # make errors web-readable
    $errmsg =~ s/\n/<BR>\n/g;

    $self->sendto_screen({title=>"Problem Encountered", content=>$errmsg});
}

# TODO: WAY too many thing still use this function.  It needs to die
#       in favor of Template Toolkit

=head2 substituteValues(CONTENT, VARIABLES, BLANK)

This is a quick and dirty substituter; template-toolkit ought to be
used instead.  This assumes a few variables or a short content.  If
the file is long and there are many variables, this routine will be
inefficient.  Maybe I should incorporate standard default vars too?

=cut

sub substituteValues {
    my $self = shift;
    $self->trace("substituteValues") if ($self->{_TRACE});
    my $content = shift || die "No content specified\n";
    my $variables = shift || return $content;
    my $blank = shift || '';
    my $proc = shift || 'none';

    my ($key, $value, $trimValue);
    my ($year, $month, $day) = ('','','');
    my ($hour, $min, $sec) = ('','','');

    my @months = ('', 'Jan', 'Feb', 'Mar', 'Apr',
                  'May', 'Jun', 'Jul', 'Aug', 'Sep',
                  'Oct', 'Nov', 'Dec');

    while (($key, $value) = each(%{$variables})) {
        if ($key) {
            $value = $blank unless ($value && $value !~ /^\s*[\.\-\_\~\?]*\s*$/);
            if ($key =~ /_date$/ && $value ne $blank) {
                $value =~ /(\d\d\d\d)-(\d\d)-(\d\d) (\d\d)\:(\d\d)\:(\d\d)/;
                ($year, $month, $day, $hour, $min, $sec) = ($1,$2,$3,$4,$5,$6);
                $day = (defined $day)? $day+0 : 0;
                if ($month) {
                    $month = $months[$month];
                }
                $content =~ s/\[\%\s*$key.date\s*\%\]/$month&nbsp;$day,&nbsp;$year/gi;
                $content =~ s/\[\%\s*$key.time\s*\%\]/$hour:$min:$sec/gi;
                $content =~ s/\[\%\s*$key\[date\]\s*\%\]/$month&nbsp;$day,&nbsp;$year/gi;
                $content =~ s/\[\%\s*$key\[time\]\s*\%\]/$hour:$min:$sec/gi;
                $content =~ s/\[\%\s*$key\s*\%\]/$value/gi;
            } elsif ($key =~ /_time$/ && $value ne $blank) {
                $value =~ /[\w\-]*\s*(\d\d):(\d\d):(\d\d)/;
                my ($hour, $min, $sec) = ($1, $2, $3);
                if (!$hour) { $hour = 0; }
                if (!$min)  { $min  = 0; }
                if (!$sec)  { $sec  = 0; }
                my $ampm = 'am';
                if ($hour > 12) {
                    $hour -= 12;
                    $ampm = 'pm';
                }
                $content =~ s/\[\%\s*$key\s*\%\]/$hour:$min $ampm/gi;
	    } elsif ($proc ne 'tt2') {
		$content =~ s/\[\%\s*$key\s*\%\]/$value/gi;
	    }

            if (length $value > 40) {
                $trimValue = sprintf "%.37s...", $value;
                $content =~ s/\[\%\s*short\s+$key\s+(\w+)\s*\%\]/$1/gi;
                $content =~ s/\[\%\s*trim\s+$key\s*\%\]/$trimValue/gi;
            } else {
                $content =~ s/\[\%\s*short\s+$key\s+(\w+)\s*\%\]/$value/gi;
                $content =~ s/\[\%\s*trim\s+$key\s*\%\]/$value/gi;
            }
        }
    }

    # Many templates need updating to work properly with tt2, so for now
    # we only run things through tt2 if specifically directed to
    if ($proc eq 'tt2') {
	warn "Processing with tt2\n" if ($self->{_DEBUG}>3);
	my $new_content = '';

	# Run content through template-toolkit
	$self->{_template}->process(\$content, $variables, \$new_content)
	    || die trace_tree()."Template process failed on content:  ".$self->{_template}->error()."\n";

	return $new_content;
    } else {
	return $content;
    }
}


########################################################################
## USERS SUBSYSTEM                                                    ##
########################################################################

=head1 Users subsystem

=head2 do_reset_password

Checks permissions, resets password and e-mails the user.

=cut
sub do_reset_password {
    my ($self, %args) = @_;
    $self->trace("do_reset_password") if ($self->{_TRACE});

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $DB          = $self->{_DB};
    my $on          = $args{'on'};
    my $dataHandle  = $args{'dataHandle'};
    my $emails      = $args{'emails'};
    my $nextCommand = $args{'nextCommand'};
    my $nextQuery   = $args{'nextQuery'};
    my $user        = $args{'user'};
    my $noticeMsg   = $args{'noticeMsg'};

    my $id = $on;
    # Consider wanting to reset a range of passwords...

    if ($user->{Permissions}->{admin} ne 'Y' && $id != $user->{$self->{_idField}}) {
        die trace_tree()."You cannot reset other people's passwords.";
    }

    # Generate a new password
    my $passwd = generateRandomPassword();

    # Prepare email msg for user
    $dataHandle->{Userpassword} = $passwd;
    my $message = loadFile("$self->{_CONFIG}->{applicationTemplates}/users/password.reset.email.tmpl");
    $message = $self->substituteValues($message, $dataHandle);

    my $sth = $DB->selectRecords($table, "$self->{_idField}=$id", '');
    my $data = $sth->fetchrow_hashref;
    if ($data && $data->{Real_email} && $data->{Real_email} ne '') {
        # Encrypt password
        my $salt = sprintf("\$1\$%.8s\$", generateRandomPassword());
        $dataHandle->{Userpassword} = md5_hex($passwd);
        $dataHandle->{Unixpassword} = crypt($passwd, $salt);
        $dataHandle->{NTpassword}   = nthash($passwd);

        # Validate the inputs and store the data
        $dataHandle->{$self->{_idField}} = $id;
        $DB->updateRecord($modulename, $id,
            $self->{_modules}->{$modulename}->{tables}, 
            %$dataHandle);
        $message = $self->substituteValues($message, $data);
        $message =~ s/\[\%\s*(\w+)\s*\%\]//g;
        $self->sendto_email($data->{Real_email}, "OSDL User Account", $message);

        my $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/users/display.reset_performed.page.tmpl");
	$self->sendto_screen({title=>ucfirst("Password has been reset"),
                         content=>$content, %$data});
    } else {
        my $content = loadFile("$self->{_CONFIG}->{applicationTemplates}/users/display.reset_failed.page.tmpl");

	$self->sendto_screen({title=>ucfirst("Could not find user's email address"), 
		      content=>$content, %$data});
    }

    return ($nextCommand, $nextQuery, $noticeMsg);
}

=head2 do_edit_password

Displays a password-editing HTML form.

=cut
sub do_edit_password {
    my ($self, %args) = @_;
    $self->trace("do_edit_password") if ($self->{_TRACE});

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $DB          = $self->{_DB};
    my $on          = $args{'on'};
    my $dataHandle  = $args{'dataHandle'};
    my $emails      = $args{'emails'};
    my $nextCommand = $args{'nextCommand'};
    my $nextQuery   = $args{'nextQuery'};
    my $user        = $args{'user'};
    my $noticeMsg   = $args{'noticeMsg'};

    my $id = $on;  # need to incorporate numbering a little more appropriately

    my $sname = "$self->{_CONFIG}->{applicationTemplates}/$modulename/edit.$table.password.form.tmpl";
    my $content = loadFile($sname);

    $self->sendto_screen({title=>"Change password", content=>$content, 
                         idField => $id,
                         on => $on,
                         %{$self->{_storageError}}});

    return ($nextCommand, $nextQuery, $noticeMsg);
}

=head2 do_store_password

Checks permissions and stores the password from the edit form.

=cut
sub do_store_password {
    my ($self, %args) = @_;
    $self->trace("do_store_password") if ($self->{_TRACE});

    my $DB=$self->{_DB};

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $on          = $args{'on'};
    my $dataHandle  = $args{'data'};
    my $user        = $args{'user'};
    my $nextCommand = $args{'nextCommand'};
    my $nextQuery   = $args{'nextQuery'};
    my $noticeMsg   = $args{'noticeMsg'};

    $nextCommand = 'display';

    $self->authorizeUser($on, $self->{_ARGS}->{auth_user}) ||
        die trace_tree()."You do not have access to change this password.";

    my $user_id = $user->{user_id};

    my $oldpasswd = $dataHandle->{Old_password};
    my $passwd = $dataHandle->{New_password};
    my $passwd2 = $dataHandle->{New_password_2};

    print "old: $oldpasswd<BR>\n" if ($self->{_TRACE});
    print "1: $passwd<BR>\n" if ($self->{_TRACE});
    print "2: $passwd2<BR>\n" if ($self->{_TRACE});

    if (!defined($passwd) || $passwd eq '') {
        $self->{_storageError}->{New_Password_Error} = 'Please enter new password';
        $command = 'reedit_password';
    }
    if (!defined($passwd2) || $passwd2 eq '') {
        $self->{_storageError}->{New_Password_Error} = 'Please enter new password twice';
        $command = 'reedit_password';
    }
    if ($passwd ne $passwd2) {
        # Error, passwords don't match
        $self->{_storageError}->{New_Password_Error} = "Passwords don't match...";
        $command = 'reedit_password';
    }

    print "want: ".$user->{Userpassword}."<BR>\n" if ($self->{_TRACE});
    print "got ".md5_hex($oldpasswd)."<BR>\n" if ($self->{_TRACE});

    if (md5_hex($oldpasswd) ne $user->{Userpassword}) {
        # Error, wrong password provided
        $self->{_storageError}->{Old_Password_Error} = "Incorrect password...";
        $command = 'reedit_password';
    }

    if ($command eq 'reedit_password') {
        return ($command, $nextQuery, $noticeMsg);
    } else {
        my $salt = sprintf("\$1\$%.8s\$", generateRandomPassword());

        # Encrypt password
        $user->{Userpassword} = md5_hex($passwd);
        $user->{Unixpassword} = crypt($passwd, $salt);
        $user->{NTpassword}   = nthash($passwd);

        # Validate the inputs and store the data
        my %data = $DB->updateRecord($modulename, $user_id,
            $self->{_modules}->{$modulename}->{tables},
            %$user);

        return ($nextCommand, $nextQuery, "Updated password");
    }
}

=head2 do_users_thanks

Displays a message thanking user for registering

=cut
sub do_users_thanks {
    my ($self, %args) = @_;
    $self->trace("do_users_thanks") if ($self->{_TRACE});
    my $DB=$self->{_DB};

    my $modulename  = $args{'modulename'};
    my $command     = $args{'command'};
    my $table       = $args{'table'};
    my $on          = $args{'on'};
    my $dataHandle  = $args{'data'};
    my $nextCommand = $args{'nextCommand'};
    my $nextQuery   = $args{'nextQuery'};
    my $defaults    = $args{'defaults'};
    my $noticeMsg   = $args{'noticeMsg'};

    my $content;
    my $filename = "$self->{_CONFIG}->{applicationTemplates}/$modulename/display.thanks.page.tmpl";
    if (-e $filename) {
        $content = loadFile($filename);
    } else {
        $content = "Thanks!  Your application has been received.";
    }

    $self->sendto_screen({title=>"You are now signed up as a user", 
		  content=>$content});
    return ($nextCommand, $nextQuery, $noticeMsg);
}

=head2 getID

Returns the configured ID field name

=cut

sub getID
{
    my ($self)=@_;

    return $self->{_idField};
}


=head2 registerQuery(MODULE, QUERY)

Registers a query either for a specific module, or if MODULE is undef,
a global query.

  $e->registerQuery(modulename => 'users',
                    queries => { query1 => [ 'table', '', '' ], });

=cut

# TODO: validate modulename
sub registerQuery {
    my ($self, %args) = @_;
    $self->trace("registerQuery");

    my $hash;
    my $loc;

    my $modulename=$args{'modulename'};
    if (!defined($modulename)) {
        # Global Query
        $loc  = "Global";
        $hash = \%{$self->{_queries}};
    } else {
        # Module-specific Query
        $loc  = "Module '$modulename'";
        $hash = \%{$self->{_modules}->{$modulename}->{queries}};
    }

    foreach my $query_name (keys %{$args{'queries'}}) {
        warn "Query $loc: $query_name\n" if ($self->{_DEBUG}>2);
        $hash->{$query_name}=$args{'queries'}->{$query_name};
    }
}

=head2 getQuery(MODULE, QUERY)

Returns the SQL conditions for a preset query, or the default query for a given
module if no preset is specificed.

=cut
sub getQuery {
    my ($self, $modulename, $query) = @_;
    $self->trace("getQuery");

    warn "want '$query' query from '$modulename' module\n" if ($self->{_DEBUG}>2);

    my $hash;
    
    $hash=\%{$self->{_modules}->{$modulename}->{queries}};
    if (exists $hash->{$query}) {
        warn "found module-specific\n" if ($self->{_DEBUG}>2);
        return @{ $hash->{$query} };
    }

    $hash=\%{$self->{_queries}};
    if (exists $hash->{$query}) {
        warn "found global\n" if ($self->{_DEBUG}>2);
        return @{ $hash->{$query} };
    }

    warn "missing\n" if ($self->{_DEBUG}>2);
    return ('', '', '');
}

=head2 makeHtmlTable(DB, MODULE, COMMAND, QUERY, STYLE, FILTER, SORT, MAXITEMS, DEFAULTS)

Inserts the specified data into an HTML table and returns it.

=cut
sub makeHtmlTable {
    my ($self, $modulename, $command, $query, $style, $addFilter, $addSort, $maxitems, $defaults) = @_;
    $self->trace("makeHtmlTable") if ($self->{_TRACE});
    my $DB = $self->{_DB};

    my ($table, $filter, $sort) = $self->getQuery($modulename, $query);
    if ($table eq '') { 
        $table = $query; 
        if ($addFilter) { $filter = $addFilter; }
    } else {
        $filter =~ s/\[\%\s*user_id\s*\%\]/$defaults->{user_id}/g;
        if ($addFilter) { $filter = join(' AND ', $filter, $addFilter); }
    }
    if ($addSort) { $sort = $addSort; }
    my $pth = "$self->{_CONFIG}->{applicationTemplates}/$modulename/$command.$query";
    my $rowTmpl = loadFile("$pth.row.tmpl");
    my $content = '';

    # Load data from database
    my $sth = $DB->selectRecords($table, $filter, $sort);

    # Insert into the template text
    my $count = 0;
    my $color = '';
    while ($sth != 0 && (my $data = $sth->fetchrow_hashref) && $count < $maxitems) {
        if ($color eq "#FFFFFF") {
            $color = "#CCCCCC";
        } else {
            $color = "#FFFFFF";
        }

        # Load the template for describing a row in a table, & fill it in
        my $line = $self->substituteValues($rowTmpl, $data, '&nbsp;');
        $line =~ s/\[\%\s*row_color\s*\%\]/$color/g;
        while ($line =~ /\[\% command (\w+) \%\]/) {
            my $cmd = $self->cgiCmd($modulename, $1, (on => $data->{$self->{_idField}}));
            my $action = qq|<A HREF="$cmd">$1</A>|;
            $line =~ s/\[\% command (\w+) \%\]/$action/;
        }
        while ($line =~ /\[\%\s*(\w+)\.(\w+)\s*\%\]/) {
            my $text = $self->getFieldText($data, $1, $2) || '';
            
            $line =~ s/\[\%\s*$1\.$2\s*\%\]/$text/g;
        }
        if ($content eq '') {
            $content = loadFile("$pth.header.tmpl");
        }
        $content .= $line;
        $count++;
    }

    return $content;
}


=head2 genericHtmlForm(DB, modulename, command, DATA)

This routine creates an HTML form using the data hash provided.
The data hash must be of the following form:

my %data = ( fields => [ ],
             descriptions => { },
             types => { },
             values => { },
             notes => { }
);

The fields array specifies the sort order of the items in the form.

Form element names will be prepended with 'd_' so that Eidetic can
parse the returned output.

Adapted from code written originally by Adam Wendt <billyjoeray@netzero.net>.

=cut
sub genericHtmlForm {
    my ($self, $modulename, $command, %data) = @_;
    $self->trace("genericHtmlForm") if ($self->{_TRACE});
    my $DB=$self->{_DB};
    my $text;

    if (! $modulename) { $modulename = ''; }

    my $rec_id = $data{values}->{$self->{_idField}} || -1;
    $text = qq|<FORM ACTION="$self->{_ARGS}->{my_url}" METHOD="POST">\n|;
    $text .= qq|<INPUT TYPE="hidden" NAME="modulename" VALUE="$modulename">\n|;
    $text .= qq|<INPUT TYPE="hidden" NAME="command" VALUE="store">\n|;
    if ($rec_id > 0) {
        $text .= qq|<INPUT TYPE="hidden" NAME="on" VALUE="$rec_id">\n|;
        $text .= qq|<INPUT TYPE="hidden" NAME="$self->{_idField}" VALUE="$rec_id">\n|;
    }
    $text .= qq|<TABLE BORDER="0">\n|;

    foreach my $field (@{$data{fields}}) {
        my $type  = $data{types}        ? $data{types}->{$field}        : 'hidden';
        my $value = $data{values}       ? $data{values}->{$field}       : '';
        my $note  = $data{notes}        ? $data{notes}->{$field}        : '';
        my $desc  = $data{descriptions} ? $data{descriptions}->{$field} : '';
        my %foreign_key;

        if (!$type)  { $type = 'string'; }
        if (!$value) { $value = ''; }

        # Check for relation into another table
        # (A foreign key, if mysql supported it)
        if ($type =~ /^int/i && $field =~ /^(\w+)\_uid/i) {
            $type = 'foreign_key';
            $desc = ucfirst($1);
            $desc =~ s/\_/ /g;
        }
        # Any fields noted as auto_increment are definitely not enterable,
        # so there is zero reason to show them anywhere on the form; skip 'em.
        next if ($note =~ /auto_increment/i);
        next if (!$field);
        next if ($field =~ /password/); # Skip fields that might be passwords

        $text .= qq| <TR>\n  <TD VALIGN="top">\n|;
        if ($desc) {
            $text .= qq|<b>$desc</b>\n|;
        } else {
            $text .= qq|&nbsp;\n|;
        }
        $text .= qq|  </TD>\n  <TD>\n|;

        # SQL Type conversion (for flexibility...  this is somewhat hackish though
        # and probably should be made much cleaner and clearer, with a deliberate
        # SQL type to Widget type converter.)
        if ($type eq 'foreign_key') {
            my $table;
            if ($field =~ /^(\w+)\_uid$/) {
                $table = lc($1);
            } elsif ($field =~ /^(\w+)\_uid\_(\w+)$/) {
                $table = lc($2);
            }
            next if ($table eq 'user');
            # Next lookup the values for this table.
            #my $dbname = "eidetic";
            #my $dbh = getDbHandle($dbname);
            if ($DB->checkTableExistance($table)) {
                my $sth = $DB->selectRecords($table, '', '');
                while (my $data = $sth->fetchrow_hashref) {
                    $foreign_key{$data->{$self->{_idField}}} = $data->{descriptor};
                }
                $type = 'menu(KEYED)';
            } else {
                # The indicated uid hasn't got an associated table
                $type = 'string';
            }
        } elsif ($type =~ /^varchar/i || $type =~ /^char/i) {
            $type = 'string';
        } elsif ($type =~ /^longtext/i) {
            $type = 'memo';
        } elsif ($type =~ /^tinyint/i) {
            # This field is probably a select...
            $type = 'string';
        } elsif ($type =~ /^int/i || $type =~ /^smallint/i || $type =~ /^numeric/i) {
            $type = 'string';
        } elsif ($type =~ /^float/i || $type =~ /^real/i || $type =~ /^double/i) {
            $type = 'string';
        } elsif ($type =~ /date/i || $type =~ /time/i) {
            $type = 'string';
        } elsif ($type =~ /^text$/i) {
            $type = 'memo';
        } elsif ($type =~ /text/i) {
            $type = 'string';
        } elsif ($type =~ /^enum/i) {
            $type =~ s/enum/radio/i;
            $type =~ s/[\'\"]//g;
        } elsif ($type =~ /boolean/i) {
            $type = "checkbox($field)";
        }

        if ($type eq 'string') {
            $text .= qq|<input type="text" name="d_$field" value="$value" SIZE="30"><br>\n|;
        } elsif ($type eq 'memo') {
            # TODO:  Add filters for if $value contains a textarea tag.
            $text .= qq|<textarea name="d_$field" cols=80 rows=10 wrap=virtual>|;
            $text .= qq|$value</textarea><br>\n|;
        } elsif ($type =~ /checkbox\((.*)\)/) {
            $text .= qq|<input type="checkbox" name="d_$field"|;
            if($value) { $text .= " checked"; }
            $text .= qq|> $1<br>\n|;
        } elsif ($type eq 'submit') {
            $text .= qq|<input typ="$type" name="d_$field"|;
            if ($value ne "") {
                $text .= qq| value="$value"|;
            }
            $text .= "><br>\n";
        } elsif ($type eq 'reset') {
            $text .= qq|<input type="$type"> $note<br>\n|;
        } elsif ($type eq 'file') {
            $text .= qq|<input type="$type" name="d_$field" value="$value"><br>\n|;
        } elsif ($type =~ /^radio\((.*)\)$/){
            foreach my $radio (split(/\,/,$1)) {
                $text .= qq|<input type="radio" name="d_$field" value="$radio"|;
                if ($value eq $radio) {
                    $text .= " checked";
                }
                $text .= qq|> $radio\n|;
            }
        } elsif ($type =~ /^menu\((.*)\)$/) {
            if (%foreign_key) {
                $text .= qq|<select name="d_$field"|;
                $text .= qq|>\n|;
                if ($1 eq 'KEYED') {
                    while ( my ($val, $val_nm) = each %foreign_key) {
                        $text .= qq|<option value="$val"|;
                        if ($value eq $val) { $text .= qq| selected|; }
                        $text .= qq|>$val_nm</option>\n|;
                    }
                } else {
                    foreach my $item (split(/\,/,$1)) {
                        $text .= qq|<option|;
                        if ($value eq $item) { $text .= qq| selected|; }
                        $text .= qq|>$item</option>\n|;
                    }
                }
                $text .= qq|</select>\n$note<br>\n|;
            } else {
                $text .= qq|<I>No items to select from</I>\n|;
            }
        } elsif ($type =~ /^image\((.*)\)$/) {
            $text .= qq|<input type="$type" name="d_$field" src="$1" alt="$value"><br\n|;
        } elsif ($type eq 'password') {
            $text .= qq|<input type="$type" name="d_$field"><br>\n|;
        } elsif ($type eq 'hidden') {
            $text .= qq|<input type="$type" name="d_$field" value="$value"><br>\n|;
        }
        if ($type ne 'hidden') {
            if ($note) { $text .= qq|$note<br>\n|; }
        } else {
            $text .= "\n";
        }
        $text .= qq|  </TD>\n </TR>\n|;
    }
    $text .= qq|</TABLE>\n|;
    $text .= qq|<INPUT TYPE="HIDDEN" NAME="command" VALUE="$command">\n|;
    $text .= qq|<P><INPUT TYPE="SUBMIT" NAME="op" VALUE="Send"></P>\n\n|;

    $text .= "</FORM>\n";

    return $text;
}


=head2 collectData

This routine gathers all of the cgi parameters that begin with the tag
'd_' (these indicate 'data' parameters that Eidetic should process.)

=cut
sub collectData {
    my ($self) = @_;
    $self->trace("collectData") if ($self->{_TRACE});
    my $cgi=$self->{_cgi};
    my (%data, $key);

    foreach $key ($cgi->param()) {
        # Ignore all keys except those starting with "d_"
        if ($key =~ /^d\_(.+)/) {
            $key=$1;
	    #don't repeat the regex
            #$key =~ s/^d\_(.+)/$1/;
            $data{$key} = $cgi->param("d_$key");
            #warn "processed $key => $data{$key}\n";
            unless ($cgi->param("command") eq 'search') {
                $cgi->param(-name=>"d_$key", 
                            -value=>'');
            }
        } elsif ($key =~ /^$self->{_idField}/) {
            $data{$key} = $cgi->param("$key");
        }
    }
    return %data;
}

=head2 cgiCmd

Creates a URL that contains the current cgi GET options for the program
ready for use inside an anchor tag or whatever.

=cut
sub cgiCmd {
    my ($self, $modulename, $command, %overrides) = @_;
    $self->trace("cgiCmd") if ($self->{_TRACE});
    my ($cmd);

    my %params = ();
    foreach my $param ($self->{_cgi}->param()) {
        if ($self->{_cgi}->param($param)
            && $self->{_cgi}->param($param) ne ''
            && $param ne 'op'
            ) {
            $params{$param} = $self->{_cgi}->param($param);
        }
    }

    foreach my $param (keys %overrides) {
        $params{$param} = $overrides{$param};
    }

    $cmd = qq|$self->{_ARGS}->{my_url}?command=$command&modulename=$modulename|;
    foreach my $param (keys %params) {
        next if ($param eq 'command' || $param eq 'modulename');
        if (! exists $self->{_eideticDefaults}->{$param} 
            || $params{$param} ne $self->{_eideticDefaults}->{$param}) {
            $cmd .= qq|&$param=$params{$param}|;
        }
    }

    return $cmd;
}

=head2 validateModule(MODULENAME)

Returns false and displays an error message if $modulename doesn\'t exist; 
returns true if it does.

=cut
sub validateModule {
    my ($self, $modulename) = @_;
    $self->trace("validateModule") if ($self->{_TRACE});
    
    die trace_tree()."Modulename undefined\n"
        if (!defined($modulename));

    if ($modulename !~ /^\w+$/) {
        if ($modulename =~ /^(\w+)/) {
            # Handles a bug encountered once where info after the URL
            # was tacked on by Apache::AuthTicket.  We solve it by
            # deliberately ignoring anything after the wonky char's
            $modulename = $1;
        } else {
            # Won't print out the modulename 'cause god knows what
            # the user's trying to do.
            die "Strange modulename parameter.\n";
        }
    }

    if (! exists $self->{_modules}->{$modulename}) {
        my $errmsg = qq|
            <P>The modulename you have provided, "$modulename" is not
            a valid modulename.</P><P>  If you reached this error message
            by clicking on an otherwise valid hyperlink, please email
            <A HREF="mailto:$self->{_CONFIG}->{webmaster}">$self->{_CONFIG}->{webmaster}</A> to let the
                maintainers know.|;
	    $self->sendto_screen({title=>"Modulename error", content=>$errmsg});
        return 0;
    } else {
        return 1;
    }
    return 1;
}

=head2 authenticate_user(DB, FILTER)

Selects the user, from DB, given by FILTER, and returns their row as a hash.

=cut
sub authenticate_user {
    my ($self, $filter) = @_;
    $self->trace("authenticate_user") if ($self->{_TRACE});

    my $DB = $self->{_AUTH_DB} || die "Could not get auth database\n";

    # If logged in, get the user's info
    my $user;
    if ($filter && $DB->isInDatabase('user', $filter)) {
        my $sth = $DB->selectRecords('user', $filter, '');
        if ($sth != 0) {
            $user = $sth->fetchrow_hashref;
        }
        
        # Get the permissions associated with this user, too
        my %permission;
        $sth = $DB->selectRecords('permission', "user_uid=".$user->{$self->{_idField}}, '');
        while (my $row = $sth->fetchrow_hashref) {
            my $resource_tag = $row->{resource_tag};
            if ($row->{access} eq 'Y' && $row->{rsf} >0) {
                $permission{$resource_tag} = 'Y';
            } else {
                $permission{$resource_tag} = 'N';
            }
        }
        $user->{Permissions} = \%permission;
    }
    return $user;
}

############
### MAIN ###
############

sub main {

    my ($self,%args)=@_;
    $self->trace("main") if ($self->{_TRACE});

    my $modulename = $args{'modulename'} || $self->{_ARGS}->{'modulename'};
    my $DB = $self->{_AUTH_DB};

    eval {

        die "No DB object defined\n" unless (defined($DB));

	if (defined $self->{_cgi}->server_software
	    && $self->{_cgi}->server_software ne 'cmdline' 
            && defined $self->{_ARGS}->{command}
	    && $self->{_ARGS}->{command} ne 'redirect') {
	    print $self->{_cgi}->header();
	}

        # Validate
        # Check if the module name is valid
        $self->validateModule($modulename) || exit(1);

        # Collect data from form
        my (%input_data) = $self->collectData();

        # If logged in, get the user's info
        my $auth_user = $self->{_ARGS}->{auth_user};
        if ($auth_user) {
            $self->{_user} = $self->authenticate_user("descriptor='$auth_user'");
        }

        # Use dispatcher to activate relevant function
        $self->dispatch(modulename => $modulename,
                        command    => $self->{_ARGS}->{command},
                        on         => $self->{_ARGS}->{on},
                        user       => $self->{_user},
                        input_data => \%input_data);

    }; # End of Eval

    $self->handleError($modulename, $@);
    $self->trace_dump();
}

1;
# vi:set ai ts=4 sw=4 expandtab:
