=head1 NAME

Eidetic::DBObject

=head1 SYNOPSIS

The basic database object operation class.  Used to get/set fields
for a specific database entry.

=head1 DESCRIPTION

=head1 FUNCTIONS

=cut

package Eidetic::DBObject;

use strict;
use Carp   qw(confess);
use vars   qw($VERSION);

# Sneaky enforced hash fields care of the Perl OO book
use fields qw(_DB _table _keyfield _dirty _attached _data);
# _DB = Eidetic::Database object
# _table = which table we're using
# _keyfield = detected unique key field to use
# _dirty = field-keyed hash of whether data in _data is flushed to DB
# _attached = defined to unique key value once attached to a record
# _data = field-keyed hash of field values once attached to a record
use vars   qw( %FIELDS );




=head2 new( Eidetic::Database $DB, scalar $table )

Using the $DB, the named $table will be looked up, and used for record
processing.

=cut

sub new {
    my ($this, $DB, $table) = @_;
    my $class = ref($this) || $this;
    my $self = bless [\%FIELDS], $class;
    
    $self->{_DB}      =$DB;
    $self->{_table}   =$table;
    $self->{_attached}=undef;

    confess "No database connection" unless (defined($self->{_DB}));

    # Look up the unique key for this table
    my %tablekeys = %{ $self->{_DB}->getTableInfo($table, 'KEYS') };

    # TODO: make this not mysql-only
    # This records all the field names, and finds the primary key
    ($self->{_keyfield})=grep($self->{_dirty}{$_}=1 &&
                         $tablekeys{$_}=~/^PRI$/,keys %tablekeys);

    confess "Could not find unique key for table '$table'"
        unless (defined($self->{_keyfield}));

    #print STDERR "key for '$table' is '$self->{_keyfield}'\n";

    return $self;
}

=head2 attach( [$key | %fields] )

If given a single $key, an existing record is looked up.

=cut

sub attach {
    my $self = shift;

    confess "No database connection" unless (defined($self->{_DB}));

    if (defined($_[0]) && scalar(@_)==1){
        # we are attaching to an existing record
        my ($id) = @_;

        confess "No database table defined" unless (defined($self->{_table}));

        # TODO: if already attached, and fields are dirty, should we flush them?
        if (defined($self->{_attached}))
        {
            # wipe out the data
            delete $self->{_data};
        }

        my $ref=$self->{_DB}->hashedQuery("SELECT * FROM $self->{_table} WHERE $self->{_keyfield}=?",
                                        $self->{_keyfield},
                                        $id);

        #confess "No matching $self->{_table} record for $self->{_keyfield} '$id'"
        # Something blew up
        return undef
            unless (defined($ref));
        # No records found
        return undef
            unless (scalar(%{ $ref }));

        #grep(print("$_ => $ref->{$_}\n"),keys %{ $ref });

        # Pull all the fields from the database
        foreach my $field (keys %{$self->{_dirty}})
        {
            # got it from the DB: not dirty
            $self->{_dirty}{$field}=0;
            $self->{_data}{$field} =$ref->{$id}->{$field};
            #print STDERR "$field => $self->{_data}{$field}\n";
        }

        $self->{_attached}=$id;
    }
    else
    {
        # TODO: we are creating a new record with any of the supplied fields
        my %args;
        %args=@_ if (defined($_[0]));

        my @fields;
        my @values;
        foreach my $field (keys %args) {
            next unless ($field);
            warn "field: $field\n";
            push(@fields,$field);
            push(@values,$args{$field});
        }
        my $fields=join(", ",@fields);
        my $values=join(", ",@values);

        my $rows=$self->{_DB}->affectedQuery("INSERT INTO $self->{_table} ($fields) VALUES ($values)",
                                    @values);
        if ($rows>0) {
            $self->{_attached}=$self->{_DB}->insertID();
        }
        else {
            return undef;
        }
    }
}

=head2 get( $field )

Returns the information for the specific field, when attached.

=cut

sub get {
    my $self  = shift;
    my $field = shift;

    confess "No such field '$field' in table '$self->{_table}'"
        unless (defined($self->{_dirty}{$field}));

    confess "Not attached" unless (defined($self->{_attached}));

    return $self->{_data}{$field};
}

=head2 set( $field, $value )

Immediately sets a field to a specific value.

=cut

sub set {
    my $self = shift;
    my ($field,$value) = @_;

    # make sure the field exists
    confess "No such field '$field' in table '$self->{_table}'"
        unless (defined($self->{_dirty}{$field}));

    # Don't let someone change the key field
    confess "Cannot set value of key field '$field' for table '$self->{_table}'"
        if ($self->{_keyfield} eq $field);

    confess "Not attached" unless (defined($self->{_attached}));

    # Is this value already set?
    return 1 if ($self->{_data}{$field} eq $value);

    # Update the value
    $self->{_data}{$field} =$value;
    $self->{_dirty}{$field}=1; # value has changed, it's dirty now

    # I'd like to have this:
    #   $self->flush() if ($self->{_autoflush});
    # But I haven't do it yet, so we'll just flush immediately instead:
    my $rows=$self->{_DB}->affectedQuery("UPDATE $self->{_table} SET $field=? WHERE $self->{_keyfield}=?",
                                    $value,
                                    $self->{_attached});
    $rows=undef if ($rows==0);
    return $rows;
}

=head2 fields

Returns an array of available fields for the object.

=cut

sub fields {
    my $self = shift;

    confess "No database connection" unless (defined($self->{_DB}));

    # Grab the table info!
    return @{ $self->{_DB}->getTableInfo($self->{_table},'FIELDS') };
}

=head2 lookForIDs(DB, WHERE/FIELDS)

Returns a list of match IDs from the Eidetic::Database, using
field/value pairs passed in the FIELDS hash, or single WHERE clause.

=cut

sub lookForUIDs
{
    my $self = shift;
    my $db = shift;
    my %query;
    my $query;
    my @values;

    my $KEY  =$self->getKey();
    my $TABLE=$self->getTable();

    confess "No database connection" unless (defined($db));

    if ((scalar(@_) % 2)==1){
        # odd arg means direct "WHERE" statement
        $query="SELECT $KEY FROM $TABLE WHERE ".shift;
        @values=@_;
    }
    else {
        %query=@_;

        my @fields;
        foreach my $field (keys %query)
        {
            push(@fields,"$field = ?");
            push(@values,$query{$field});
        }
        if (@fields) {
            $query="SELECT $KEY FROM $TABLE WHERE ".join(" AND ",@fields);
        }
        else {
            $query="SELECT $KEY FROM $TABLE";
        }
    }

    #print $query," :$KEY: ";
    #grep(print("$_ "),@values);
    #print "\n";

    my $ref=$db->hashedQuery($query,$KEY,@values);

    # did something blow up?
    return ()
        unless (defined($ref));

    return sort keys %{ $ref };
}



1;
# vi:set ai ts=4 sw=4 expandtab:
