=head1 NAME

Eidetic::Database

=head1 SYNOPSIS

Eidetic's database routines

=head1 DESCRIPTION

=head1 FUNCTIONS

=cut

package Eidetic::Database;

use strict;
use Carp                 qw(confess);
use vars qw($VERSION @ISA %DEFAULTS %CONFIG);

use DBI                  qw();
use Eidetic::Utility     qw(parseOn trim);
# Sneaky enforced hash fields care of the Perl OO book
use fields qw(_brand _name _user _host _pass _handle _tables);
use vars qw( %FIELDS );


use constant DEBUGGING => 0;

require Exporter;
require DynaLoader;

our @ISA = qw(Exporter DynaLoader);
our @EXPORT_OK = ( );

my $idField = 'uid';

my $getfrom_db_count = 0;
my $getfrom_db_firstId = 0;
my $getfrom_db_lastId = 0;
my $getfrom_db_maxitems = 25;

=head2 new(CONFIG_HASH)

The CONFIG_REF hash should include any of the following parameters.
Defaults are noted:

brand  => Brand of database to use (Default is 'mysql')

For the DBI:mysql type, the following must also be defined:
name   => Name of database (required: no default)
host   => Host of the database (required: no default)
user   => Username to use for access (required: no default)
pass   => Password for the user (required: no default)

For example:

$hash{name}="stuff";
$hash{host}="localhost";
$hash{user}="stuff_user";
$hash{pass}="secret";

$db=new Eidetic::Database(%hash);

=cut


sub new {
    my ($this, %args) = @_;
    my $class = ref($this) || $this;
    my $self = bless [\%FIELDS], $class;
    
    # bring in the args
    while (my($field, $value) = each %args)
    {
      confess "Invalid argument '$field'"
        unless (exists $FIELDS{"_$field"});
      $self->{"_$field"}=$value;

    }

    # handle a DBI-style database connection
    $self->{_brand} ||= "mysql";

    if ($self->{_brand} eq "mysql")
    {
        confess "Missing required DBI::mysql parameter: name"
          unless (defined($self->{_name}));
        confess "Missing required DBI::mysql parameter: host"
          unless (defined($self->{_host}));
        confess "Missing required DBI::mysql parameter: user"
          unless (defined($self->{_user}));
        confess "Missing required DBI::mysql parameter: password"
          unless (defined($self->{_pass}));

        $self->{_handle} = DBI->connect(join(":",
                                              "DBI",
                                              $self->{_brand},
                                              $self->{_name},
                                              $self->{_host}),
                                         $self->{_user},
                                         $self->{_pass});
        confess "Failed to connect to database: $DBI::errstr\n"
          unless (defined($self->{_handle}));
    }
    else
    {
        confess "Unknown database brand '$self->{_brand}'";
    }

    return $self;
}

=head2 getfrom_db(TABLE, START, FILTER, SORT, FIRST, QTY)

Gets specified rows from database, using selectRecords, and returns a handle
for the first row found.

=cut

sub getfrom_db {
    my $self=shift;
    my ($table, $on, $filter, $sort, $first, $qty) = @_;

    my $rightify = ($on =~ /\!/);
    my $newFilter = parseOn($on);

    if ($filter) {
        $filter .= " AND ";
    }
    $filter .= "$idField $newFilter";

    my $handle = $self->selectRecords($table, $filter, $sort, $first, $qty);

    if ($handle == 0) {
        # No available records were found
        confess "No available records were found, so we couldn't process the request.";
    }

    $getfrom_db_count = 0;
    if ($rightify && $handle->rows > $getfrom_db_maxitems) {
        $getfrom_db_count = $getfrom_db_maxitems - $handle->rows;
    }
    while ($getfrom_db_count<0) {
        # throw away unwanted initial elements
        my $data = $handle->fetchrow_hashref;
        $getfrom_db_count++;
    }

    return $handle;
}

=head2 getnext_db(HANDLE)

Takes $handle returned from getfrom_db and returns the next row or C<undef>

=cut

# Returns a reference to a hash containing the data item
sub getnext_db($) {
    my $self=shift; 
    my ($handle) = @_;

    $getfrom_db_count++;

    my $data = $handle->fetchrow_hashref;
    return $data;
}

#=head2 prepareDatabases(MODULENAME)
#
#Given a valid module name, this initializes all the databases that it
#will need.
#
#=cut
#sub prepareDatabases {
#    my $self=shift; 
#    my ($modulename, @dbs) = @_;
#
#    # Prepare the databases
#    my $dbname;
#    # Connect to databases as required
#    foreach $dbname (@dbs) {
#        # This is a little kludgy and needs cleaning up
#        if ($dbname ne '') {
#            $db_handles{$dbname} = DBI->connect(@{$db_info{$dbname}});
#            if (! $db_handles{$dbname}) {
#                die trace_tree()."failure Could not connect to database $dbname";
#                # TODO:  Add better error handling
#            }
#        }
#    }
#}
#
#=head2 getDbHandle(DBNAME)
#
#Returns a handle to db DBNAME; throws on failure.
#
#=cut
#sub getDbHandle($) {
#    my $self=shift; 
#    my ($dbname) = @_;
#    if (exists $db_handles{$dbname}) {
#        return $db_handles{$dbname};
#    } elsif (exists $db_info{$dbname}) {
#        $db_handles{$dbname} = DBI->connect(@{$db_info{$dbname}});
#        if (! $db_handles{$dbname}) {
#            die trace_tree()."failure Could not connect to database $dbname.  \n"
#		. "This can occur if the database is not running, or \n"
#		. "if the wrong db user/pass/host info is specified in \n"
#		. "the config file.\n";
#        }
#        return $db_handles{$dbname};
#    } else {
#        die trace_tree()."Unknown database $dbname.  This is most likely a bug in \n"
#	    . "the perl code for the application.  Please report this bug.\n";
#    }
#}

=head2 createDatabaseTable(TABLE, SCHEMA)

Creates $table in database, with fields and types given by:
$schema{$field}=$type.  Probably {mysql,pg}-specific?

=cut

sub createDatabaseTable {
    my $self=shift; 
    my ($table, %schema) = @_;

    if (! $self->{_handle}) {
        confess "Not connected to database!";
    }

    if ($table eq '') {
        confess "Invalid table name '$table'";
    }

    my $cmd = "CREATE TABLE $table (";

    my $sep = '';
    while (my ($field, $type) = each %schema) {
        $cmd .= "$sep $field $type";
        $sep = ',';
    }
    $cmd .= ')';

    # Execute the SQL statement 
    my $sth = $self->{_handle}->prepare($cmd)
        or confess "failure Cannot prepare database command: " . $self->{_handle}->errstr;
    my $rows_affected = $sth->execute();

    # TODO:  Perhaps also create an INDEX?
}

=head2 addDatabaseFields(TABLE, SCHEMA)

Like createTable but doesn't create anything.' DB-specific?

=cut

sub addDatabaseFields {
    my $self=shift; 
    my ($table, %schema);

    # TODO: verify the field doesn't already exist, and is a valid field name

    my $cmd = "ALTER TABLE $table ADD (";

    my $sep = '';
    while (my ($field, $type) = each %schema) {
        $cmd .= "$sep $field $type";
        $sep = ',';
    }
    $cmd .= ')';

    # Execute the SQL statement  - should I use do?
    my $sth = $self->{_handle}->prepare($cmd)
        or confess "failure Cannot prepare database command: " . $self->{_handle}->errstr;
    my $rows_affected = $sth->execute();
}

#=head2 closeDatabases( )
#
#This routine simply closes whatever databases are currently opened.
#
#=cut
#sub closeDatabases() {
#    my $self=shift; 
#    my $dbname;
#    # We're done with the db, so disconnect it
#    foreach $dbname (keys %db_handles) {
#        $db_handles{$dbname}->disconnect();
#    }
#}

=head2 DESTROY() Destructor
=cut

sub DESTROY
{
    my $self=shift;
    if (defined($self->{_handle}))
    {
        $self->{_handle}->disconnect();
        $self->{_handle}=undef;
    }
}


=head2 checkMissingElements

Given a table name and reference to a data hash
this routine fills in defaults and returns a list of
missing elements.

=cut

sub checkMissingElements {
    my $self=shift; 
    my ($table, $data) = @_;
    my @missingElements;
    my $schemaDefault = $self->getTableInfo($table, 'DEFAULTS');

    if (! $self->{_handle}) {
        confess "No database connection!";
    }

    if (! $table) {
        confess "Internal error in checkMissing Elements:  $table not defined";
    }

    if (! $data) {
        confess "Internal error in checkMissing Elements:  $data not defined";
    }

    foreach my $key (keys %{ $schemaDefault }) {
        if (! exists $data->{$key}) {
            if (! $schemaDefault->{$key}) {
                $data->{$key} = '';
            } elsif ($schemaDefault->{$key} eq 'required') {
                push (@missingElements, $key);
            } else {
                $data->{$key} = $schemaDefault->{$key};
            }
        }
    }
    return @missingElements;
}


=head2 checkInvalidElements($$$)

Returns empty list

=cut

sub checkInvalidElements($$$) {
    my $self=shift; 
    my @invalidElements;
    return @invalidElements;
}


=head2 insertRecord(TABLE, DATA)

insertRecord - Inserts data into a db table, assuming it meets schema

=cut

sub insertRecord {
    my $self=shift; 
    my ($table, %data) = @_;
    my ($i);

    # For this table, determine what the columns are
    my @fields = @{ $self->getTableInfo($table, 'STORE_FIELDS') };

    # Assemble the SQL statement
    my $cmd = "INSERT INTO $table (" . join(',', @fields) . ") VALUES (";

    my %type = %{ $self->getTableInfo($table, 'TYPES') };
    for ($i = 0; $i < ($#fields); $i++) {
        $cmd .= $self->formatItem($data{$fields[$i]}, $type{$fields[$i]}) . ",";
    }
    $cmd .= $self->formatItem($data{$fields[$#fields]}, $type{$fields[$#fields]}) . ")";

    # Execute the SQL statement
    my $sth = $self->{_handle}->prepare($cmd)
        or confess "failure Cannot prepare database command: " . $self->{_handle}->errstr;
    my $rows_affected = $sth->execute();

    if ($rows_affected == 0) {
        confess "Insert failed for $cmd $!: " . $self->{_handle}->errstr;
    }

    # retrieve any autoincrement that the database created

    if ($sth->{mysql_insertid}) {
        return $sth->{mysql_insertid};
    } else {
        return $sth->{insertid};
    }
}


sub changePermission {
    my $self=shift; 
    my ($user_uid, $resource_tag, $access) = @_;

    my $cmd = 'INSERT INTO permission ';
    $cmd .=   '(user_uid, resource_tag, access, trust_score, rsf)';
    $cmd .=   "VALUES ($user_uid, '$resource_tag', '$access', 100, 1)";

    my $sth = $self->{_handle}->prepare($cmd)
        or confess "failure Cannot prepare database command: " . $self->{_handle}->errstr;

    return $sth->execute();
}

=head2 isInDatabase(TABLE, FILTER)

This routine searches to see if any records match the given filter in the
given table within the database handle provided.  It returns
true if there are one or more records matching the criteria.

=cut
sub isInDatabase {
    my $self=shift; 
    my ($table, $filter) = @_;

    $filter="" unless (defined($filter));
    confess "No table given for filter '$filter'" unless $table;
    my @fields = @{ $self->getTableInfo($table, 'FIELDS') };

    my $cmd = "SELECT $fields[0] FROM $table";

    # Need WHERE condition based on $filter, and need to do sorting
    if ($filter ne '') {
        $cmd .= " WHERE $filter";  # For now we'll use plain old SQL
    }

    my $sth = $self->{_handle}->prepare($cmd)
        or confess "failure Cannot prepare database command '$cmd':  " . $self->{_handle}->errstr();
    my $numRows = $sth->execute;
    if ($sth->errstr()) {
        confess "failure Cannot execute database command '$cmd': " . $sth->errstr();
    } elsif (! $numRows) {
        confess "failure Cannot execute database command '$cmd'";
    }

    return $numRows>0;
}

=head2 quickQuery(QUERY, [ARGS...])

Performs a quick query and returns the result from fetchrow to the caller.

=cut

sub quickQuery
{
    my $self=shift;
    my $query=shift;
    my @args=shift;

    confess "No database connection" unless (defined($self->{_handle}));

    my $sth=$self->{_handle}->prepare($query);
    $sth->execute(@args);
    return $sth->fetchrow;
}


=head2 getId(TABLE, DESCRIPTOR)

This is a convenience routine which does a quick lookup in the db to 
retrieve the uid matching a given descriptor, or -1 if the descriptor
cannot be found, or 0 if it was blank.

=cut
sub getId {
    my $self=shift; 
    my ($table, $descriptor) = @_;

    $descriptor =~ s/^\s+//;
    $descriptor =~ s/\s+$//;
    return 0 if $descriptor eq '';

    my $cmd = "SELECT $idField FROM $table where descriptor='$descriptor'";

    my $sth = $self->{_handle}->prepare($cmd)
        or confess "failure Cannot prepare database command '$cmd': " . $self->{_handle}->errstr();
    my $numRows = $sth->execute;
    if ($sth->errstr()) {
        confess "failure Cannot execute database command '$cmd': " . $sth->errstr();
    }

    if ($numRows < 1) { return -1; }

    my ($id) = $sth->fetchrow_array;
    return $id;
}

=head2 selectRecords(TABLE, FILTER, SORT, FIRST, QTY)

Begins a query of the database using the specified filter and sort,
and returns the query statement handle.  Use $queryHandle->fetchall_arrayref,
$queryHandle->fetchrow_hashref, or etc. as desired to get the data.

=cut

sub selectRecords {
    my $self=shift; 
    my ($table, $filter, $sort, $first, $qty) = @_;

    $filter="" unless (defined($filter));
    $sort  ="" unless (defined($sort));
    confess "selectRecords called with no table specified, for filter='$filter', sort='$sort'" unless $table;
    my @fields;
    my $junction = '';
    my $left = 0;

    $first= 0 if (!defined($first));
    $qty  = 0 if (!defined($qty));

    # This figures out junction table joining
    ($table, $junction) = split /\<\-/, $table;

    if (defined($junction) && $junction ne '') {
        if ($junction =~ /^\+(.*)$/) {
            $junction = $1;
            $left = 1;
        }
        my @junction_fields = @{ $self->getTableInfo($junction, 'FIELDS') };
        my @target_fields = @{ $self->getTableInfo($table, 'FIELDS') };
        foreach my $target_field (@target_fields) {
            push @fields, "$table.$target_field";
        }
        foreach my $junction_field (@junction_fields) {
            push @fields, "$junction.$junction_field";
        }
    } else {
        @fields = @{ $self->getTableInfo($table, 'FIELDS') };
    }
    my $cmd = "SELECT " . join(', ', @fields) . " FROM $table";

    # Need WHERE condition based on $filter, and need to do sorting
    if ($left) {
        $cmd .= " LEFT JOIN $junction ON $junction.";
        $cmd .= ucfirst ($table);
        $cmd .= "_$idField = $table.$idField";

    } elsif ($junction) {
        $cmd .= ", $junction";
        $cmd .= " WHERE $junction.rsf<>0 AND $table.rsf<>0 AND $junction.";
	if ($junction =~ /resource/) {
	    $cmd .= 'Resource';
	} else {
	    $cmd .= ucfirst ($table);
	}
        $cmd .= "_$idField = $table.$idField";
    } else {
        $cmd .= " WHERE rsf<>0";
    }
    if ($filter ne '') {
        $cmd .= " AND ($filter)";
    }

    # Add sorting, if $sort is actually in @fields
    if ($sort =~ /\!/) {
        my $s = $sort; $s =~ s/\!//g;
        if (grep { $_ eq $s } @fields) {
    	    $cmd .= " ORDER BY $s DESC";  # For now we'll use plain old SQL
    	} else {
    	    print STDERR "$s does not appear to be in @fields\n" if (DEBUGGING > 1);
    	}
    } elsif ($sort ne '') {
    	if (grep { $_ eq $sort } @fields) {
    	    $cmd .= " ORDER BY $sort";
    	    print STDERR "Ordering by $sort\n" if (DEBUGGING > 1);
    	}
        elsif (grep { $_ eq "$table.$sort" } @fields) {
    	    $cmd .= " ORDER BY $table.$sort";
    	}
        elsif (my @matches = grep { $sort =~ /$_/ } @fields) {
    	    $cmd .= " ORDER BY $matches[0]";
    	    print STDERR "Found @matches\n" if (DEBUGGING > 1);
    	    print STDERR "Ordering by $matches[0]\n" if (DEBUGGING > 1);
    	}
        else {
    	    print STDERR "Err, sorting by unknown field $sort\n" if (DEBUGGING > 1);
    	    print STDERR "Couldn't find it in @fields\n" if (DEBUGGING > 1);
    	}
    }
    elsif ($junction) {
        # By default, we sort by the sub-table's uid
        $cmd .= " ORDER BY $idField";
    }

    if ($first>0 && $qty>0) {
        $cmd .= " LIMIT $first,$qty";
    }
    my $sth = $self->{_handle}->prepare($cmd)
        or confess "failure Cannot prepare database command '$cmd': " . $self->{_handle}->errstr();
    my $numRows = $sth->execute
        or confess "failure Cannot execute database command '$cmd': " . $sth->errstr();

    return $sth;
}

=head2  getDatabases

Returns a list of available databases in the db server

=cut

sub getDatabases {
    my $self=shift; 
    my ($server) = @_;
    my @databases = $self->{_handle}->func($server, '_ListDBs');

    if (not @databases) {
        # emit error
        confess "No databases found";
    }
    return @databases;
}


=head2  getTables

Returns a list of available tables in a database

=cut
sub getTables {
    my $self=shift; 
    my ($server) = @_;
    my @tables = $self->{_handle}->func($server, '_ListTables');

    if (not @tables) {
        # emit error
        confess "No tables found";
    }
    return @tables;
}

=head2 checkTableExistance(TABLE)

Returns true if the table exists, false if not.

=cut

sub checkTableExistance {
    my $self=shift; 
    my ($table) = @_;
    my $cmd = "SHOW FIELDS FROM $table";
    my $sth = $self->{_handle}->prepare($cmd)
        or confess "failure Cannot prepare database statement: " . $self->{_handle}->errstr;
    if ($sth->execute()) {
        return 1;
    } else {
        return 0;
    }
}


=head2 getTableInfo(TABLE, INFOTYPE)

This routine retrieves a data structure which describes a particular table
and is a real workhorse for Eidetic.  It returns a reference to an array or
hash for the $infotype.

It requires and assumes that the table actually
exists in the database, and if it does not, it dies.
Thus call checkTableExistance($table) to verify first

=cut

sub getTableInfo {
    my $self=shift;
    my ($table, $infotype) = @_;

    confess "Table name not specified for '$infotype'" unless ($table ne '');

    if (! $self->{_handle}) {
        confess "Database not active for reading $table";
    }

    if (! exists $self->{_tables}{$table}) {
        my $cmd = '';

        if ($self->{_brand} eq 'mysql') {
            $cmd = "SHOW FIELDS FROM $table";
        } elsif ($self->{_brand} eq 'postgresql') {

# The code below was contributed by Malcolm Walker to add Postgres support for
# this routine, in around 2001, when Eidetic was successfully ported to Postgres.
# Since then, though, Postgres support has not been maintained and is unlikely
# to work right out of the box.  In any case, this code is being left in place
# to help anyone wishing to explore porting of the db to postgres in the future.
# This code enables Postgres to return a list of fields for a given table, in
# much the same way as the mysql command SHOW FIELDS.  This is an architectural
# requirement of Eidetic, and it is vital that this dynamic field lookup be done
# in order for Eidetic's auto-discovery mechanisms to work properly.

            $cmd = qq| SELECT
   a.attname,
   a.attrelid,
   format_type(a.atttypid, a.atttypmod),
   a.attnotnull,
   a.atthasdef,
   (
     SELECT 'PRI' from pg_index i, pg_attribute a2
     WHERE c.relname = '$table'
       AND i.indrelid = c.oid
       AND i.indexrelid = a2.attrelid
       AND a.attrelid = c.oid
       AND a2.attname = a.attname
   ),
   (
     SELECT substring(d.adsrc for 128)
     FROM pg_attrdef d, pg_class c
     WHERE c.relname = '$table'
       AND c.oid = d.adrelid
       AND d.adnum = 1
       AND a.atthasdef = 't'
   )
   FROM pg_attrdef d, pg_class c, pg_attribute a
   WHERE c.relname = '$table'
     AND a.attnum > 0
     AND a.attrelid = c.oid
     AND d.adrelid = c.oid
 |;

        } else {
             confess "Unsupported database brand";
        }

        my $sth = $self->{_handle}->prepare($cmd)
            or confess "Cannot prepare database statement ($cmd): " . $self->{_handle}->errstr;
        my $out = $sth->execute()
            or confess "Cannot execute database statement ($cmd): " . $sth->errstr;

        my (%descs, %types, %nulls, %keys, %defaults, %extra, %access);
        my (@fields, @input_fields, @output_fields, @store_fields);
        while (my @table_info = $sth->fetchrow_array) {
            my ($field, $type, $null, $key, $default, $extra);

            if ($self->{_brand} eq 'mysql') {
                ($field, $type, $null, $key, $default, $extra) = @table_info;
            } elsif ($self->{_brand} eq 'postgresql') {
                my $foo;
                ($field, $foo, $type, $null, $key, $default) = @table_info;
            } else {
                confess "Unsupported brand of database";
            }

            push @fields, $field;

            $descs{$field}    = ucfirst($field);
            $descs{$field}    =~ s/\_(\w)/ \U$1/g; # Convert _ to ' ' & title-case
            $types{$field}    = $type;
            $nulls{$field}    = $null;
            $keys{$field}     = $key;
            $defaults{$field} = $default;
            $extra{$field}    = $extra;

            if ($field ne $idField) {
                push @store_fields, $field;
            }
            if ($field ne $idField && $field ne 'rsf' &&
                $field !~ /^created/ && $field !~ /^last\_updated/) {
                push @input_fields, $field;
            }
            push @output_fields, $field;
        }

        $self->{_tables}{$table} = {
#            PERMISSIONS   => \%permissions,
            RESOURCE_NAME => '',
            FIELDS        => \@fields,
            STORE_FIELDS  => \@store_fields,
            INPUT_FIELDS  => \@input_fields,
            OUTPUT_FIELDS => \@output_fields,
            FIELD_ACCESS  => \%access,
            DESCS         => \%descs,
            TYPES         => \%types,
            NULLS         => \%nulls,
            KEYS          => \%keys,
            DEFAULTS      => \%defaults,
            EXTRA         => \%extra
            };
        # RESOURCE_NAME is used to group tables that are part of a
        # system.
    }

    if (exists $self->{_tables}{$table}->{$infotype}) {
        return $self->{_tables}{$table}->{$infotype};
    } else {
        confess "failure Could not get '$infotype' for '$table'";
    }
}


=head2 checkType(TYPE, ITEM)

This routine verifies that the given item is of the type specified

=cut

sub checkType {
    my $self=shift; 
    my ($type, $item) = @_;

    # Later, should I change this into a hash of types to fptrs,
    # where each function validates a particular data type, so
    # users of this function can add custom type checking?
    if (!  ($type)
        || ($type =~ /^tinyint/i
            && $item =~ /^[+-]?\d+$/)
        || ($type =~ /^int/i
            && $item =~ /^[+-]?\d+$/)
        || ($type =~ /^float/i
            && $item =~ /^([+-]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))?$/)
        || ($type =~ /^varchar\((\d+)\)$/i
            && length $item < $1)
        || ($type =~ /^date/i
            && $item =~ /^\d\d\d\d-\d\d-\d\d$/)
        || ($type eq 'text')
        || ($type eq 'string')
        )
        # TODO:  tinyint, date, text
    {
        return 1;
    } else {
        return "Bad type $type specified.";
    }
}


=head2 formatItem(ITEM, FORMAT)

Puts data item into the format specified

=cut

sub formatItem {
    my $self=shift; 
    my ($item, $format) = @_;

    if (! defined $item) {
        $item = '';
    } else {
	$item = trim($item);
	$item =~ s/\"/\\\"/g;
	$item =~ s/\'/\\\'/g;
    }
    if (! defined $format) {
        $format = '';
    }

    # Later I should add a more sophisticated formatting system
    # Perhaps do a lookup of the format from the database?
    # I'd need to load the database with default formats, of course...
    if ($format eq 'string' || $format =~ /char/i || $format =~ /text/i) {
        return "'" . $item . "'";
    } elsif ($format =~ /enum/) {
        return "'" . $item . "'";
    } elsif ($format eq 'date' || $format eq 'datetime') {
        return "'" . $item . "'";
    } elsif ($item eq '') {
        return '0';
    } else {
        return $item;
    }
}

=head2 validateAndStoreRecord(MODULE, USERID, DATA)

#Validates and stores %data. Doesn't check for duplicate or invalide
elements.

=cut

sub validateAndStoreRecord {
    my $self=shift; 
    my ($modulename, $user_id, $tables, %data) = @_;

    my ($dbt, $table, $dbname, @missingElements, @invalidElements, $rec_id);
    my @dbts = @{$tables};

    if ($modulename eq 'tests' || $modulename eq 'host_types') {
        push @dbts, 'eidetic:host_type_to_test_junction';
        push @dbts, 'eidetic:host_type_to_test';
    }
    if ($modulename eq 'projects' || $modulename eq 'hosts') {
        push @dbts, 'eidetic:project_to_host_junction';
        push @dbts, 'eidetic:project_to_host';
    }
    if ($modulename eq 'projects' || $modulename eq 'users') {
        push @dbts, 'eidetic:project_to_user_junction';
        push @dbts, 'eidetic:project_to_person';
    }
    foreach $dbt (@dbts) {
        ($dbname, $table) = split /:/, $dbt;

        # If this table has a uniqueness field, make sure we're not adding a duplicate
        next unless $self->checkTableExistance($table);
        my %keys = %{ $self->getTableInfo($table, 'KEYS') };

        my $filter = '';
        my $keyfield;
        foreach $keyfield (keys %keys) {
            next if (! $keyfield || !exists $keys{$keyfield} || !exists $data{$keyfield});
            if ($keys{$keyfield} eq 'MUL' || $keys{$keyfield} eq 'PRI') {
                $filter = "$keyfield='$data{$keyfield}'"; # assuming string here - could be anything
                if ($self->isInDatabase($table, $filter)) {
                    # TODO:
                    # Indicate to the user that this record already appears to be in there
                }
            }
        }

        my $sth = $self->{_handle}->prepare("SELECT NOW()");
        $sth->execute;
        my $now = $sth->fetchrow;

        # Add the housekeeping data elements
        $data{created_by} = $user_id;
        $data{created_date} = $now;
        $data{last_updated_by} = $user_id;
        $data{last_updated_date} = $now;
        $data{rsf} = 1;

        # Verify we have the required elements; if not, make a list.
        # This also fills in defaults where appropriate.
        push (@missingElements, $self->checkMissingElements($table, \%data));

        # Next verify the data is in the right formats, and do conversions as appropriate
        push (@invalidElements, $self->checkInvalidElements($table, \%data));

        # Update junction tables...
        if ($table =~ /_to_/) {
            $self->updateJunctions($modulename, $table, %data);
        } else {
            $data{$idField} = $self->insertRecord($table, %data);
        }
    }

    return %data;
}



=head2 updateJunctions (MODULE, TABLE, DATA)

Update junction tables.

=cut

sub updateJunctions {
    my $self=shift; 
    my ($modulename, $table, %data) = @_;
    my $id = $data{$idField};

    # This indicates a junction table
    my ($table1, $table2) = split /\_to\_/, $table;
    $table2 =~ s/\_junction$//;
    
    my ($source, $target) = ('', '');
    if ($modulename eq $table1."s") {
        $source = "$table1\_$idField";
        $target = "$table2\_$idField";
    } elsif ($modulename eq $table2."s") {
        $target = "$table1\_$idField";
        $source = "$table2\_$idField";
    } else {
	    confess "Could not figure out source and target from $modulename, $table1, and $table2";
    }

    my %elements;
    # Extract all the elements that need adding to the db
    while (my ($field, $value) = each %data) {
        if ($field =~ /^$table\:\:$target\[/) {
            $elements{$value} = 1;  # Mark it
        }
    }

    my $cmd = "SELECT $source, $target, rsf FROM $table WHERE $source = $id";
    my $sth = $self->{_handle}->prepare($cmd)
        or confess "failure Cannot prepare database command: " . $self->{_handle}->errstr;
    $sth->execute;
    my (@inserts, @deletes, @undeletes);
    while (my $data = $sth->fetchrow_hashref) {
        if (! exists $elements{$data->{$target}}) {
            # Mark this one for deletion
            push @deletes, $data->{$target};
            delete $elements{$data->{$target}};
        } elsif ($elements{$data->{$target}} == 1) {
            # undelete if necessary
            if ($data->{rsf} < 1) {
                push @undeletes, $data->{$target};
                # Else we need do nothing to it
            }
            delete $elements{$data->{$target}};
        }
    }

    # If any elements are left, INSERT them
    @inserts = keys %elements;

    my $item;
    # Next perform the database commands to do all of these updates
    foreach $item (@inserts) {
        $cmd = "INSERT INTO $table ($source, $target, rsf) ";
        $cmd .= "VALUES ($id, $item, 1)";

        $sth = $self->{_handle}->prepare($cmd)
            or confess "failure Cannot prepare database command: " . $self->{_handle}->errstr;
        $sth->execute;
    }

    foreach $item (@deletes) {
        $cmd = "DELETE FROM $table WHERE ";
        $cmd .= "$source=$id AND $target=$item ";

        $sth = $self->{_handle}->prepare($cmd)
            or confess ."failure Cannot prepare database command: " . $self->{_handle}->errstr;
        $sth->execute;
    }

    foreach $item (@undeletes) {
        $cmd = "UPDATE $table SET rsf=1 WHERE ";
        $cmd .= "$source=$id AND $target=$item ";

        $sth = $self->{_handle}->prepare($cmd)
            or confess "failure Cannot prepare database command: " . $self->{_handle}->errstr;
        $sth->execute;
    }
}


=head2 updateRecord(MODULE, USER, DATA)

#Updates $module's tables with %data.

=cut

sub updateRecord {
    my $self=shift; 
    my ($modulename, $user_id, $tables, %data) = @_;

    my $id = $data{$idField};
    if ($id !~ /^\d+$/) {
        confess "$idField '$id' not specified correctly in updateRecord";
    }

    my ($dbt, $table, $dbname, @missingElements, $rec_id);
    my @dbts = @{$tables};
    if ($modulename eq 'tests' || $modulename eq 'host_types') {
        push @dbts, "eidetic:test_to_host_type";
    }
    foreach $dbt (@dbts) {
        ($dbname, $table) = split /\:/, $dbt;

        my %type = %{ $self->getTableInfo($table, 'TYPES') };
        my $sep = '';

        if ($table =~ /_to_/) {
            $self->updateJunctions($modulename, $table, %data);

        } else {
            my $sth = $self->{_handle}->prepare("SELECT NOW()");
            $sth->execute;
            my $now = $sth->fetchrow;

            # Add the housekeeping data elements
            $data{last_updated_by} = $user_id;
            $data{last_updated_date} = $now;

            my $cmd = '';

            foreach my $field (keys %data) {
                if ($field eq $idField) {
                    # skip it.
                } elsif (exists $type{$field} && exists $data{$field}) {
                    if ($cmd eq '') { $cmd = "UPDATE $table SET "; }
                    $cmd .= $sep . $field . '=' . $self->formatItem($data{$field}, $type{$field});
                    $sep = ', ';
                } else {
                    # User requested a change to a field that we know nothing about
                }
            }

            if ($cmd ne '') {
                $cmd .= " WHERE $idField=$id";
                # Execute the SQL statement  - should I use do?
                $sth = $self->{_handle}->prepare($cmd)
                    or confess "failure Cannot prepare database command: " . $self->{_handle}->errstr;
                my $rows_affected = $sth->execute();

                if ($rows_affected == 0) {
                    confess "Update failed for $cmd $!: " . $sth->errstr;
                }
            }
        }
    }
    return %data;
}

=head2 hashedQuery(QUERY, KEY, [ARGS...])

Performs a query and returns the keyed result from fetchall_hashref.

=cut

sub hashedQuery
{
    my $self=shift;
    my $query=shift;
    my $key=shift;

    confess "No database connection" unless (defined($self->{_handle}));

    my $sth=$self->{_handle}->prepare($query);
    $sth->execute(@_);
    my $ref=$sth->fetchall_hashref($key);

    # spit an error, if one exists
    return $sth->err || $ref;
}

=head2 affectedQuery(QUERY, [ARGS...])

Performs a query and returns the number of affected rows.

=cut

sub affectedQuery
{
    my $self=shift;
    my $query=shift;

    confess "No database connection" unless (defined($self->{_handle}));

    my $sth=$self->{_handle}->prepare($query);
    $sth->execute(@_);
    my $rows=$sth->rows;

    # spit an error, if one exists
    return $sth->err || $rows;
}

=head2 insertID()

Returns the ID of the last insert

=cut

sub insertID
{
    my $self=shift;

    confess "No database connection" unless (defined($self->{_handle}));

    # TODO unmysql-ize this
    return $self->{_handle}->{'mysql_insertid'};
}

1;
# vi:set ai ts=4 sw=4 expandtab:
