#!/usr/bin/perl

=head1 NAME

Eidetic::Utility

=head1 SYNOPSIS

Various utility routines for Eidetic

=head1 DESCRIPTION

=head1 FUNCTIONS

=cut

package Eidetic::Utility;

use strict;
use vars qw($VERSION @ISA %DEFAULTS %CONFIG);
use vars qw();

use Fcntl qw(:DEFAULT :flock);

use constant DEBUGGING => 0;

require Exporter;
require DynaLoader;

our @ISA = qw(Exporter DynaLoader);
our @EXPORT_OK = qw( 
		     sendto_file
		     generateId
		     incrementId
		     generateRandomPassword
		     stringify
		     cleanFilename
		     checkCriteria
		     parseOn
		     trace_tree
		     );
our @EXPORT = qw(loadFile trim);

my $idField = 'uid';

=head2 trim(OUT)

Removes leading and trailing ws from @out.

=cut
# Removes leading and trailing whitespace
sub trim {
    my @out = @_;

    for (@out) {
        s/^\s+//;
        s/\s+$//;
    }
    return wantarray ? @out : $out[0];
}

=head2 loadFile(FNAME)

Loads $fname, returns all lines as a string. On failure dies; if $fname
doesn\'t exist gives a wacky HTML stack trace.

=cut
sub loadFile {
    my ($fname) = @_;
    $fname = "" unless (defined($fname));
    if (! -e $fname) {
        die trace_tree()."$fname does not appear to exist."
    }
    open(GF_FILE, "< $fname") || die trace_tree()."$fname could not be opened";
    my @lines = <GF_FILE>;
    close(GF_FILE) || die trace_tree()."$fname could not be closed";
    return join '', @lines;
}



=head2 sendto_file(FNAME, TITLE, CONTENT)

Saves text to a file.

=cut
sub sendto_file {
    my ($fname, $title, $content) = @_;
    $content="" unless (defined($content)); # avoid warnings

    open FILE, "> $fname" || die trace_tree()."Failed to open $fname";
    if ($title) {
        print FILE "$title";
    }
    print FILE $content || die trace_tree()."Failed to write to $fname";
    close FILE || die trace_tree()."Failed to close $fname";
    return 1;
}


=head2 generateId(FNAME)

    Generates a random id; uses $fname to find which id\'s are taken. Adds id to
$fname.

=cut
sub generateId {
    my ($fname) = @_;
    # Load file to find which ID's are taken
    my $num=0;

    while ($num == 0) {
        # Pick a random number
        $num = int rand(1000000);

        if (-e $fname) {
            open FILE, "< $fname" or die trace_tree()."Could not generate new Id for form";

            while (<FILE>) {
                chomp;
                if ($num == $_) {
                    $num = 0;
                    last;
                }
            }
            close FILE;
        }
    }

    # This is an okay ID, so add it to the file
    if (-e $fname) {
        open FILE, ">> $fname" or die trace_tree()."Could not store new Id for form";
    } else {
        open FILE, "> $fname" or die trace_tree()."Could not store new Id for form";
    }

    flock(FILE, LOCK_EX) || die trace_tree()."Could not lock $fname: $!\n";
    print FILE "$num\n" || die trace_tree()."Could not write to $fname: $!\n";
    flock(FILE, LOCK_UN) || die trace_tree()."Could not unlock $fname: $!\n";
    close FILE || die trace_tree()."Could not close $fname: $!\n";;
    return $num;
}

=head2 incrementId(FNAME)

This is like the previous function but uses incremental ID\'s

=cut
sub incrementId {
    my ($fname) = @_;

    sysopen(FILE, $fname, O_RDWR|O_CREAT) or die trace_tree()."Can't open numfile: $!";
    flock(FILE, LOCK_EX) or die trace_tree()."Can't write-lock numfile: $!";
    my $num = <FILE> || 0;
    seek(FILE, 0, 0) or die trace_tree()."Can't rewind numfile: $!";
    truncate(FILE,0) or die trace_tree()."Can't truncate numfile: $!";
    print FILE $num+1, "\n" or die trace_tree()."Can't write numfile: $!";
    close (FILE) or die trace_tree()."Can't close numfile: $!";
    return $num+1;
}

=head2 generateRandomPassword( )

Generates a random eight character password, using letters, numbers,
and a little punctuation.

=cut
sub generateRandomPassword {
    my @chars = ( 'A' .. 'Z', 'a' .. 'z', 0 .. 9, qw(! @ ^ &) );
    return join("", @chars[ map { rand @chars } ( 1 .. 8 ) ]);
}


=head2 stringify(VALUE, LEVEL)

    Stringifies $value, handling refs correctly; array & hash refs are indented by
$level. Returns undef if it can\'t be stringified.

=cut
sub stringify {
    my ($value, $level) = @_;
    my $type = ref($value);
    if (! $type) {
        # Just a scalar
        return $value;
    } elsif ($type eq 'SCALAR') {
        # Scalar reference
        return stringify($$value);
    } elsif ($type eq 'ARRAY') {
        # Array reference
        my ($item, @items);
        my $spaces = ' ' x $level;
        foreach $item (@$value) {
            push @items, $spaces . stringify($item, $level+1);
        }
        if ($level == 0) {
            return join(",\n", @items) . "\n";
        } else {
            return "(\n$spaces" . join(",\n$spaces", @items) . "\n$spaces)\n";
        }
    } elsif ($type eq 'HASH') {
        # Hash reference
        my ($item, @items);
        my $spaces = ' ' x $level;
        while ( my ($key, $value) = each %$value) {
            my $valStr = stringify($value, $level+1);
            if ($valStr) {
                push @items, "$spaces$key = $valStr\n";
            }
        }
        if ($level == 0) {
            return join(",\n", @items) . "\n";
        } else {
            return "{\n$spaces" . join(",\n$spaces", @items) . "\n$spaces}\n";
        }
    } elsif ($type eq 'CODE') { # These we'll not do anything with for now
    } elsif ($type eq 'REF') {
    } elsif ($type eq 'GLOB') {
    } elsif ($type eq 'IO') {
    } elsif ($type eq 'LVALUE') {
    } elsif ($type eq 'Regexp') {
    }
    return undef;
}


=head2 cleanFilename(FILENAME)

Returns $filename with all invalid filename characters stripped.

=cut
sub cleanFilename {
    my (@filename) = split /\//, $_[0];
    my $filename = pop @filename;

    $filename =~ s/[\-\s\+]/_/g;# Change space, -, and + to _
	$filename =~ s/\\/\//g;     # Change \ to /
    $filename =~ s/[^\w\/]//g;  # Remove all punctuation except _ and /
    $filename =~ s/.*\///;      # Trim dir path out & remove /
    return $filename;
}


=head2 checkCriteria(A, OP, B)

Performs an indicated comparison operator on values A and B, and
returns true or false as appropriate.  Operators are:

EQ, GT, LT, GE, LE, NE --  ==, >, <, >=, <=, !=  respectively

CN -- Contains

NC -- Does not contain

ST -- Starts with

EN -- Ends with

=cut
sub checkCriteria {
    my ($a, $op, $b) = @_;

    return 0 if (!$op || !$a || !$b);

    return ( ( $op eq 'EQ' && $a eq $b) ||
             ( $op eq 'GT' && $a gt $b) ||
             ( $op eq 'LT' && $a lt $b) ||
             ( $op eq 'GE' && $a ge $b) ||
             ( $op eq 'LE' && $a le $b) ||
             ( $op eq 'NE' && $a ne $b) ||
             ( $op eq 'CN' && $a =~ /$b/) ||
             ( $op eq 'NC' && $a !~ /$b/) ||
             ( $op eq 'ST' && $a =~ /^$b/) ||
             ( $op eq 'EN' && $a =~ /$b$/) );
}



=head2 parseOn(ON)

Parses ON into SQL. Syntax: $firstId-$lastId means rows with ids between
$firstId and $lastId; if either is omitted, default is 0 or largest id,
respectively. Will also do joins eventually.

=cut

sub parseOn {
    my @a = split /\,/, $_[0];
    my (@ins, @betweens, $firstId, $lastId);
    foreach (@a) {
        if (/^\-(\d+)\!/) {
            push @betweens, " BETWEEN 0 AND $1";
        } elsif (/^(\d+)-$/) {
            $firstId = $1;
            push @betweens, " >= $firstId";
        } elsif (/\-/) {
            ($firstId, $lastId) = split /\-/;
            push @betweens, " BETWEEN $firstId AND $lastId";
        } elsif (! /\-/) {
            push @ins, $_;
        } else {
            # syntax error; ignore.
        }
    }
    my ($ins);
    if ($#ins>=0) {
        $ins = "IN (" . join(', ', @ins) . ") ";
        return join(" OR $idField ", ($ins, @betweens));
    } else {
        return join(" OR $idField ", @betweens);
    }
}

=head2 note 

Generic interface for debug notices
Provide way to select amount of info on a per-file / per-subr basis

=cut

# TODO: use confess or similar routines
my $current_subr;
sub trace_tree {
    my (@msg) = @_;
    my $prefix = '';

    # TODO:  Make htmlification optional
    my $endl = "\n";
    my $prefix_str = "  ";

    my $levels = 1;
    my $text = '';
    while ((caller($levels++))[3]) {
        $prefix .= $prefix_str;
    }
    $text .= "Trace: $endl";
    for (my $i=$levels-1; $i>0; $i--) {
	my $subr = ((caller($i))[3]) || '_top_';
	my $line = ((caller($i))[2]) || 'unknown';
	my $file = ((caller($i))[1]) || '';
	$file =~ s/^.*[^\w\.\-]([\w\.\-]+)$/$1/;

	$text .= ($prefix_str x ($levels-$i-1)) . " + " . $subr . "()";
	if ($file && $line) {
	    $text .= " [$file @ $line]";
	}
	$text .= ": $endl";
    }
    $text .= $endl;
    foreach my $msg (@msg) {
	$text .= $msg . $endl;
    }
    return $text;
}
1;
